"""
An instance of the Traveling Tournament Problem (TTP).

The Traveling Tournament Problem (TTP) describes the logistics of a sports
league. In this league, `n` teams compete. In each time slot, each team plays
against one other team. In each game, one team plays at home and one team
plays away with the other team. In each round, every team plays once against
every other team. The league may have multiple :attr:`~moptipyapps.ttp.\
instance.Instance.rounds`. If there are two rounds, then each team plays
against each other team once at home and once abroad. If a team plays at home
(or abroad) several times in a row, this is called a "streak". There are
minimum and maximum streak length constraints defined, for both at home and
abroad. Additionally, if team A plays a team B in one time slot, then the
exact inverse game cannot take place in the next time slot. A minimum
number of games must take place in between for separation. There can also be
a maximum separation length.

David Van Bulck of the Sports Scheduling Research group, part of the
Faculty of Economics and Business Administration at Ghent University, Belgium,
maintains "RobinX: An XML-driven Classification for Round-Robin Sports
Timetabling", a set of benchmark data instances and results of the TTP.
We provide some of these instances as resources here. You can also download
them directly at <https://robinxval.ugent.be/RobinX/travelRepo.php>. Also,
see <https://robinxval.ugent.be/> for more information.

1. David Van Bulck. Minimum Travel Objective Repository. *RobinX: An
   XML-driven Classification for Round-Robin Sports Timetabling.* Faculty of
   Economics and Business Administration at Ghent University, Belgium.
   https://robinxval.ugent.be/
2. Kelly Easton, George L. Nemhauser, and Michael K. Trick. The Traveling
   Tournament Problem Description and Benchmarks. In *Principles and Practice
   of Constraint Programming (CP'01),*  November 26 - December 1, 2001, Paphos,
   Cyprus, pages 580-584, Berlin/Heidelberg, Germany: Springer.
   ISBN: 978-3-540-42863-3. https://doi.org/10.1007/3-540-45578-7_43
   https://www.researchgate.net/publication/220270875
"""

from typing import Callable, Final, Iterable, TextIO, cast

import numpy as np
from defusedxml import ElementTree  # type: ignore
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_INT, int_range_to_dtype
from moptipy.utils.strings import sanitize_name
from pycommons.io.path import Path, file_path
from pycommons.types import check_int_range, check_to_int_range, type_error

from moptipyapps.tsp.instance import Instance as TSPInstance
from moptipyapps.ttp.robinx import open_resource_stream


def _from_stream(stream: TextIO) -> "Instance":
    """
    Read a TTP instance from an `robinxval.ugent.be`-formatted XML file.

    This procedure ignores most of the data in the file and only focuses on
    the instance name, the team names, and the distance matrix as well as the
    constraints for home streak length, away streak length, and same-game
    separations. Everything else is ignored.

    :param stream: the text stream
    :return: the instance
    """
    used_names: set[str] = set()
    team_names: dict[int, str] = {}
    distances: dict[tuple[int, int], int] = {}
    name: str | None = None
    rounds: int | None = None
    home_streak_min: int | None = None
    home_streak_max: int | None = None
    away_streak_min: int | None = None
    away_streak_max: int | None = None
    separation_min: int | None = None
    separation_max: int | None = None

    for _event, element in ElementTree.iterparse(stream, forbid_dtd=True,
                                                 forbid_entities=True,
                                                 forbid_external=True):
        if element.tag is None:
            continue
        tag: str = element.tag.strip().lower()
        if tag == "instancename":
            if name is None:
                name = sanitize_name(element.text).lower()
                if name in used_names:
                    raise ValueError(
                        f"name {element.text!r} invalid, as it "
                        f"maps to {name!r}, which is already used.")
                used_names.add(name)
            else:
                raise ValueError(f"already got name={name!r}, but tag "
                                 f"'InstanceName' appears again?")
        elif tag == "distance":
            t1: int = check_to_int_range(element.attrib["team1"],
                                         "team1", 0, 1_000_000)
            t2: int = check_to_int_range(element.attrib["team2"],
                                         "team2", 0, 1_000_000)
            dst: int = check_to_int_range(
                element.attrib["dist"], "dist", 0, 1_000_000_000_000)
            if t1 == t2:
                if dst == 0:
                    continue
                raise ValueError(f"distance for team1={t1}, team2={t2} is"
                                 f" {dst} but must be 0.")
            tpl: tuple[int, int] = (t1, t2)
            if tpl in distances:
                raise ValueError(
                    f"got distance={dst} for {tpl!r}, but "
                    f"{distances[tpl]} was already specified before")
            distances[tpl] = dst
        elif element.tag == "team":
            team: int = check_to_int_range(element.attrib["id"], "id",
                                           0, 1_000_000)
            nn: str = element.attrib["name"]
            tname: str = nn.strip()
            if tname in used_names:
                raise ValueError(f"name {nn!r} is invalid, as it maps to "
                                 f"{tname!r}, which is already used.")
            used_names.add(tname)
            team_names[team] = tname
        elif element.tag == "numberroundrobin":
            if rounds is not None:
                raise ValueError(f"rounds already set to {rounds}")
            rounds = check_to_int_range(element.text, "rounds", 1, 1000)
        elif tag == "ca3":
            if "mode1" not in element.attrib:
                continue
            if "mode2" not in element.attrib:
                continue
            if element.attrib["mode2"].lower() != "games":
                continue
            mode = element.attrib["mode1"].lower()
            mi = check_to_int_range(
                element.attrib["min"], "min", 0, 1_000_000) \
                if "min" in element.attrib else None
            ma = check_to_int_range(
                element.attrib["max"], "max", 1, 1_000_000) \
                if "max" in element.attrib else None
            if mode == "h":
                if mi is not None:
                    if home_streak_min is not None:
                        raise ValueError("minimum home streak already defined")
                    home_streak_min = 1 if mi < 1 else mi
                if ma is not None:
                    if home_streak_max is not None:
                        raise ValueError("maximum home streak already defined")
                    home_streak_max = ma
            elif mode == "a":
                if mi is not None:
                    if away_streak_min is not None:
                        raise ValueError("minimum away streak already defined")
                    away_streak_min = 1 if mi < 1 else mi
                if ma is not None:
                    if away_streak_max is not None:
                        raise ValueError("maximum away streak already defined")
                    away_streak_max = ma
        elif tag == "se1":
            mi = check_to_int_range(
                element.attrib["min"], "min", 0, 1_000_000) \
                if "min" in element.attrib else None
            ma = check_to_int_range(
                element.attrib["max"], "max", 0, 1_000_000) \
                if "max" in element.attrib else None
            if mi is not None:
                if separation_min is not None:
                    raise ValueError("minimum separation already defined")
                separation_min = mi
            if ma is not None:
                if separation_max is not None:
                    raise ValueError("maximum separation already defined")
                separation_max = ma

    if name is None:
        raise ValueError("did not find instance name")
    n_teams: Final[int] = len(team_names)
    if n_teams <= 0:
        raise ValueError("did not find any team name")
    if len(used_names) != (n_teams + 1):
        raise ValueError(f"set of used names {used_names!r} has wrong "
                         f"length, should be {n_teams + 1}.")
    dm: np.ndarray = np.zeros((n_teams, n_teams), DEFAULT_INT)
    for tup, dst in distances.items():
        dm[tup[0], tup[1]] = dst

    if rounds is None:
        rounds = 2
    ll: Final[int] = rounds * n_teams - 1
    if home_streak_min is None:
        home_streak_min = 1
    if home_streak_max is None:
        home_streak_max = min(max(home_streak_min, 3), ll)
    if away_streak_min is None:
        away_streak_min = 1
    if away_streak_max is None:
        away_streak_max = min(max(away_streak_min, 3), ll)
    if separation_min is None:
        separation_min = 1
    if separation_max is None:
        separation_max = min(max(separation_min, 1), ll)

    return Instance(name, dm, [team_names[i] for i in range(n_teams)],
                    rounds, home_streak_min, home_streak_max, away_streak_min,
                    away_streak_max, separation_min, separation_max)


#: The instances made available within this package are taken from
#: <https://robinxval.ugent.be/RobinX/travelRepo.php>, where the following
#: descriptions are given:
#: - *Constant Distance (`con*`):* The constant distance instances are the
#:   most simple instances in which the distance between the home venues of
#:   any two teams is one. In this case, Urrutia and Ribeiro showed that
#:   distance minimization is equivalent with break maximization.
#: - *Circular Distance (`circ*`):* Somewhat similar are the circular
#:   distance instances in which the teams' venues are placed on a
#:   circle. Any two consecutive teams are connected by an edge and the
#:   distance between two teams is equal to the minimal number of edges that
#:   must be traversed to get to the other team. Although traveling
#:   salesperson problems with a circular distance matrix have a trivial
#:   solution, it remains challenging to solve circular traveling tournament
#:   instances.
#: - *Galaxy (`gal*`):* This artificial instance class consists of the Galaxy
#:   instances that use a 3D-space that embeds the Earth and 39 other
#:   exoplanets.
#: - *National league (`nl*`):* The `nl`-instances are based on air distance
#:   between the city centers from teams in the National League of the Major
#:   League Baseball.
#: - *National football league (`nfl*`):* The NFL-instances are based on air
#:   distance between the city centers from teams in the National Football
#:   League.
#: - *Super 14 (`sup*`):* The super 14 instances are based on air distance
#:   between the city centers from teams in the Super 14 rugby cup.
#: - *Brazilian (`bra24`)):* The Brazilian instance is based on the air
#:   distance between the home cities of 24 teams in the main division of the
#:   2003 edition of the Brazilian soccer championship.
#: - *Linear (`line*`):* In the linear instances, `n` teams are located on a
#:   straight line with a distance of one unit separating each pair of
#:   adjacent teams.
#: - *Increasing distance (`incr*`):* In the increasing distance instances,
#:   `n` teams are located on a straight line with an increasing distance
#:   separating each pair of adjacent teams such that the distance between
#:   team `k` and `k+1` equals `k`.
_INSTANCES: Final[tuple[str, ...]] = (
    "bra24", "circ4", "circ6", "circ8", "circ10", "circ12", "circ14",
    "circ16", "circ18", "circ20", "circ22", "circ24", "circ26", "circ28",
    "circ30", "circ32", "circ34", "circ36", "circ38", "circ40", "con4",
    "con6", "con8", "con10", "con12", "con14", "con16", "con18", "con20",
    "con22", "con24", "con26", "con28", "con30", "con32", "con34", "con36",
    "con38", "con40", "gal4", "gal6", "gal8", "gal10", "gal12", "gal14",
    "gal16", "gal18", "gal20", "gal22", "gal24", "gal26", "gal28", "gal30",
    "gal32", "gal34", "gal36", "gal38", "gal40", "incr4", "incr6", "incr8",
    "incr10", "incr12", "incr14", "incr16", "incr18", "incr20", "incr22",
    "incr24", "incr26", "incr28", "incr30", "incr32", "incr34", "incr36",
    "incr38", "incr40", "line4", "line6", "line8", "line10", "line12",
    "line14", "line16", "line18", "line20", "line22", "line24", "line26",
    "line28", "line30", "line32", "line34", "line36", "line38", "line40",
    "nfl16", "nfl18", "nfl20", "nfl22", "nfl24", "nfl26", "nfl28", "nfl30",
    "nfl32", "nl4", "nl6", "nl8", "nl10", "nl12", "nl14", "nl16", "sup4",
    "sup6", "sup8", "sup10", "sup12", "sup14")


class Instance(TSPInstance):
    """An instance of Traveling Tournament Problem (TTP)."""

    #: the names of the teams
    teams: tuple[str, ...]
    #: the number of rounds
    rounds: int
    #: the minimum number of games that can be played at home in a row
    home_streak_min: int
    #: the maximum number of games that can be played at home in a row
    home_streak_max: int
    #: the minimum number of games that can be played away in a row
    away_streak_min: int
    #: the maximum number of games that can be played away in a row
    away_streak_max: int
    #: the minimum number of games between a repetition of a game setup
    separation_min: int
    #: the maximum number of games between a repetition of a game setup
    separation_max: int
    #: the data type to be used for plans
    game_plan_dtype: np.dtype

    def __new__(cls, name: str, matrix: np.ndarray, teams: Iterable[str],
                rounds: int, home_streak_min: int, home_streak_max: int,
                away_streak_min: int, away_streak_max: int,
                separation_min: int, separation_max: int,
                tour_length_lower_bound: int = 0) -> "Instance":
        """
        Create an instance of the Traveling Salesperson Problem.

        :param cls: the class
        :param name: the name of the instance
        :param matrix: the matrix with the data (will be copied)
        :param teams: the iterable with the team names
        :param rounds: the number of rounds
        :param tour_length_lower_bound: the lower bound of the tour length
        :param home_streak_min: the minimum number of games that can be played
            at home in a row
        :param home_streak_max: the maximum number of games that can be played
            at home in a row
        :param away_streak_min: the minimum number of games that can be played
            away in a row
        :param away_streak_max: the maximum number of games that can be played
            away in a row
        :param separation_min: the minimum number of games between a repetition
            of a game setup
        :param separation_max: the maximum number of games between a repetition
            of a game setup
        """
        names: Final[tuple[str, ...]] = tuple(map(str.strip, teams))
        n: Final[int] = len(names)
        if (n % 2) != 0:
            raise ValueError(f"the number of teams must be even, but is {n}.")
        if n != len(set(names)):
            raise ValueError(f"some team name appears twice in {teams!r} "
                             f"after fixing it to {names!r}.")
        for nn in names:
            for char in nn:
                if char.isspace():
                    raise ValueError(
                        f"team name must not contain space, but {nn} does.")

        obj: Final[Instance] = cast(Instance, super().__new__(
            cls, name, tour_length_lower_bound, matrix, rounds * n))

        if (obj.shape[0] != n) or (obj.shape[1] != n) or (obj.n_cities != n):
            raise ValueError(f"inconsistent n_teams={n}, n_cities="
                             f"{obj.n_cities} and shape={obj.shape}")
        #: the names of the teams that compete
        obj.teams = names
        #: the number of rounds
        obj.rounds = check_int_range(rounds, "rounds", 1, 100)
        ll: Final[int] = rounds * n - 1  # an upper bound for streaks
        #: the minimum number of games that can be played at home in a row
        obj.home_streak_min = check_int_range(
            home_streak_min, "home_streak_min", 1, ll)
        #: the maximum number of games that can be played at home in a row
        obj.home_streak_max = check_int_range(
            home_streak_max, "home_streak_max", home_streak_min, ll)
        #: the minimum number of games that can be played away in a row
        obj.away_streak_min = check_int_range(
            away_streak_min, "away_streak_min", 1, ll)
        #: the maximum number of games that can be played away in a row
        obj.away_streak_max = check_int_range(
            away_streak_max, "away_streak_max", away_streak_min, ll)
        #: the minimum number of games between a repetition of a game setup
        obj.separation_min = check_int_range(
            separation_min, "separation_min", 0, ll)
        #: the maximum number of games between a repetition of a game setup
        obj.separation_max = check_int_range(
            separation_max, "separation_max", separation_min, ll)
        #: the data type to be used for the game plans
        obj.game_plan_dtype = int_range_to_dtype(-n, n)
        return obj

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the instance to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("I") as kv:
        ...         Instance.from_resource("gal4").log_parameters_to(kv)
        ...     print(repr('@'.join(l.get_log())))
        'BEGIN_I@name: gal4@class: moptipyapps.ttp.instance.Instance@\
nCities: 4@tourLengthLowerBound: 67@tourLengthUpperBound: 160@symmetric: T@\
dtype: h@rounds: 2@homeStreakMin: 1@homeStreakMax: 3@\
awayStreakMin: 1@awayStreakMax: 3@separationMin: 1@separationMax: 6@\
gamePlanDtype: b@END_I'
        """
        super().log_parameters_to(logger)
        logger.key_value("rounds", self.rounds)
        logger.key_value("homeStreakMin", self.home_streak_min)
        logger.key_value("homeStreakMax", self.home_streak_max)
        logger.key_value("awayStreakMin", self.away_streak_min)
        logger.key_value("awayStreakMax", self.away_streak_max)
        logger.key_value("separationMin", self.separation_min)
        logger.key_value("separationMax", self.separation_max)
        logger.key_value("gamePlanDtype", self.game_plan_dtype.char)

    @staticmethod
    def from_file(path: str, lower_bound_getter: Callable[[
            str], int] | None = None) -> "Instance":
        """
        Read a TTP instance from a `robinX` formatted XML file.

        :param path: the path to the file
        :param lower_bound_getter: ignored
        :return: the instance

        >>> from os.path import dirname
        >>> inst = Instance.from_file(dirname(__file__) + "/robinx/con20.xml")
        >>> inst.name
        'con20'
        """
        file: Final[Path] = file_path(path)
        with file.open_for_read() as stream:
            try:
                return _from_stream(cast(TextIO, stream))
            except (TypeError, ValueError) as err:
                raise ValueError(f"error when parsing file {file!r}") from err

    @staticmethod
    def list_resources(symmetric: bool = True,
                       asymmetric: bool = True) -> tuple[str, ...]:
        """
        Get a tuple of all the TTP instances available as resource.

        All instances of the `robinX` set provided here are symmetric.

        :param symmetric: include the instances with symmetric distance
            matrices
        :param asymmetric: include the asymmetric instances with asymmetric
            distance matrices
        :return: the tuple with the instance names

        >>> len(Instance.list_resources())
        118
        >>> len(Instance.list_resources(False, True))
        0
        >>> len(Instance.list_resources(True, False))
        118
        """
        return _INSTANCES if symmetric else ()

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load a TTP instance from a resource.

        :param name: the name string
        :return: the instance

        >>> insta = Instance.from_resource("bra24")
        >>> insta.n_cities
        24
        >>> insta.name
        'bra24'
        >>> insta.teams[0]
        'Atl.Mineiro'
        >>> insta.teams[1]
        'Atl.Paranaense'
        >>> insta.rounds
        2
        >>> insta.home_streak_min
        1
        >>> insta.home_streak_max
        3
        >>> insta.away_streak_min
        1
        >>> insta.away_streak_max
        3
        >>> insta.separation_min
        1
        >>> insta.separation_max
        46
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        container: Final = Instance.from_resource
        inst_attr: Final[str] = f"__inst_{name}"
        if hasattr(container, inst_attr):  # instance loaded?
            return cast(Instance, getattr(container, inst_attr))

        with open_resource_stream(f"{name}.xml") as stream:
            inst: Final[Instance] = _from_stream(stream)

        if inst.name != name:
            raise ValueError(f"got {inst.name!r} for instance {name!r}?")
        if inst.n_cities <= 1000:
            setattr(container, inst_attr, inst)
        return inst



"""
The objective function for the Quadratic Assignment Problem.

The candidate solutions to the QAP are
:mod:`~moptipy.spaces.permutations` `p` of length `n` of the numbers `0`, `1`,
..., `n-1`. An :mod:`~moptipyapps.qap.instance` of the Quadratic Assignment
Problem (QAP) presents a `n*n` matrix `D` with
:attr:`~moptipyapps.qap.instance.Instance.distances` and a `n*n` matrix `F`
with flows :attr:`~moptipyapps.qap.instance.Instance.flows`. The objective
value, subject to minimization, is then the
`sum( F[i,j] * D[p[i], p[j]] for i, j in 0..n-1 )`, i.e., the sum of the
products of the flows between facilities and the distances of their assigned
locations.

1. Eliane Maria Loiola, Nair Maria Maia de Abreu, Paulo Oswaldo
   Boaventura-Netto, Peter Hahn, and Tania Querido. A survey for the
   Quadratic Assignment Problem. European Journal of Operational Research.
   176(2):657-690. January 2007. https://doi.org/10.1016/j.ejor.2005.09.032.
2. Rainer E. Burkard, Eranda Çela, Panos M. Pardalos, and
   Leonidas S. Pitsoulis. The Quadratic Assignment Problem. In Ding-Zhu Du,
   Panos M. Pardalos, eds., Handbook of Combinatorial Optimization,
   pages 1713-1809, 1998, Springer New York, NY, USA.
   https://doi.org/10.1007/978-1-4613-0303-9_27.

>>> inst = Instance.from_resource("bur26a")
>>> QAPObjective(inst).evaluate(np.array([
...     25, 14, 10, 6, 3, 11, 12, 1, 5, 17, 0, 4, 8, 20,
...     7, 13, 2, 19, 18, 24, 16, 9, 15, 23, 22, 21]))
5426670

>>> inst = Instance.from_resource("nug12")
>>> QAPObjective(inst).evaluate(np.array([
...     11, 6, 8, 2, 3, 7, 10, 0, 4, 5, 9, 1]))
578

>>> inst = Instance.from_resource("tai12a")
>>> QAPObjective(inst).evaluate(np.array([
...     7, 0, 5, 1, 10, 9, 2, 4, 8, 6, 11, 3]))
224416
"""


from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from pycommons.types import type_error

from moptipyapps.qap.instance import Instance
from moptipyapps.shared import SCOPE_INSTANCE


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def _evaluate(x: np.ndarray, distances: np.ndarray, flows: np.ndarray) -> int:
    """
    Evaluate a solution to the QAP.

    :param x: the permutation representing the solution
    :param distances: the distance matrix
    :param flows: the flow matrix
    :return: the objective value
    """
    result: int = 0
    for i, xi in enumerate(x):
        for j, xj in enumerate(x):
            result += flows[i, j] * distances[xi, xj]
    return int(result)


class QAPObjective(Objective):
    """An objective function for the quadratic assignment problem."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the QAP objective function.

        :param instance: the one-dimensional ordering problem.
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the instance data
        self.instance: Final[Instance] = instance

    def evaluate(self, x) -> float:
        """
        Compute the quadratic assignment problem objective value.

        :param x: the permutation of elements
        :return: the sum of flows times distances
        """
        return _evaluate(x, self.instance.distances, self.instance.flows)

    def __str__(self):
        """
        Get the name of this objective.

        :return: `"qap"`
        """
        return "qap"

    def lower_bound(self) -> float:
        """
        Get the lower bound of this objective function.

        :return: the lower bound
        """
        return self.instance.lower_bound

    def upper_bound(self) -> float:
        """
        Get the upper bound of this objective function.

        :return: the upper bound
        """
        return self.instance.upper_bound

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as scope:
            self.instance.log_parameters_to(scope)



"""
A system of three coupled oscillators.

There are three oscillators located in a two-dimensional plane. Thus, there
are a total of six coordinates.

The initial starting point of the work here were conversations with
Prof. Dr. Bernd NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute
of Technology in Shenzhen, China (哈尔滨工业大学(深圳)) as well as the
following paper:

1. Ruiying Li, Bernd R. Noack, Laurent Cordier, Jacques Borée, Eurika Kaiser,
   and Fabien Harambat. Linear genetic programming control for strongly
   nonlinear dynamics with frequency crosstalk. *Archives of Mechanics.*
   70(6):505-534. Warszawa 2018. Seventy Years of the Archives of Mechanics.
   https://doi.org/10.24423/aom.3000. Also: arXiv:1705.00367v1
   [physics.flu-dyn] 30 Apr 2017. https://arxiv.org/abs/1705.00367.
"""

from math import pi
from typing import Final

import numba  # type: ignore
import numpy as np
from pycommons.io.path import Path
from pycommons.types import check_int_range

from moptipyapps.dynamic_control.starting_points import (
    make_interesting_starting_points,
)
from moptipyapps.dynamic_control.system import System

#: pi to the square
__PI2: Final[float] = pi * pi


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __3_coupled_oscillators(state: np.ndarray, _: float,
                            control: np.ndarray, out: np.ndarray) -> None:
    """
    Compute the differential equations of a controlled 3-oscillators system.

    The system is six-dimensional. There are three oscillators, and each one
    is updated. There is one control value.
    This function implements equation (3.1) of the paper.

    :param state: the state of the system
    :param _: the time index, which is ignored
    :param control: the output of the controller
    :param out: the differential, i.e., the output of this function
    """
    a1: Final[float] = state[0]
    a2: Final[float] = state[1]
    a3: Final[float] = state[2]
    a4: Final[float] = state[3]
    a5: Final[float] = state[4]
    a6: Final[float] = state[5]

    r1sqr: Final[float] = (a1 * a1) + (a2 * a2)
    r2sqr: Final[float] = (a3 * a3) + (a4 * a4)
    r3sqr: Final[float] = (a5 * a5) + (a6 * a6)
    sigma1: Final[float] = -r1sqr + r2sqr - r3sqr
    sigma2: Final[float] = 0.1 - r2sqr
    sigma3: Final[float] = -0.1

    b: Final[float] = control[0]

    # compute the differential system
    out[0] = (sigma1 * a1) - a2
    out[1] = (sigma1 * a2) + a1
    out[2] = (sigma2 * a3) - (pi * a4)
    out[3] = (sigma2 * a4) + (pi * a3) + b
    out[4] = (sigma3 * a5) - (__PI2 * a6)
    out[5] = (sigma3 * a6) + (__PI2 * a5) + b


def _sinus_control(
        _: np.ndarray, time: float, __: float, out: np.ndarray) -> None:
    """
    Present the sinus-based control law from the paper.

    :param _: the state, ignored
    :param time: the time
    :param __: ignored
    :param out: the output destination
    """
    out[0] = 0.07 * np.sin(__PI2 * time)


def _lgpc3_36(
        state: np.ndarray, time: float, _: float, out: np.ndarray) -> None:
    """
    Present the LGPC-3 equation 3.6.

    :param state: the state
    :param time: the time
    :param _: ignored
    :param out: the output destination
    """
    out[0] = np.tanh(np.sin(np.tanh(
        3.0 * state[1] * np.sin(time) * np.sin(__PI2 * time) - state[3])))


class __TO(System):
    """The internal three oscillators class."""

    def describe_system_without_control(
            self, dest_dir: str, skip_if_exists: bool = True) \
            -> tuple[Path, ...]:
        result = list(super().describe_system_without_control(
            dest_dir, skip_if_exists))
        result.extend(self.describe_system(
            "open loop 0.07sin(\u03C0²t)", _sinus_control, 0.0,
            f"{self.name}_open_loop", dest_dir, skip_if_exists))
        result.extend(self.describe_system(
            "LGPC-3", _lgpc3_36, 0.0,
            f"{self.name}_lgpc3", dest_dir, skip_if_exists))
        return tuple(result)


def make_3_couple_oscillators(n_points: int) -> System:
    """
    Create the oscillator system.

    :param n_points: the number of training points
    :return: the Lorenz system
    """
    check_int_range(n_points, "n_points", 1, 1_000)
    tests: Final[np.ndarray] = \
        np.array([[0.1, 0, 0.1, 0, 0.1, 0]], float)
    # similar to training = make_interesting_starting_points(111, tests)
    training: Final[np.ndarray] = np.array([[
        -0.00721145, -0.00399194, -0.01078522, 0.0795534, 0.02929588,
        0.01588136], [
        0.13202113, 0.05538969, -0.01011213, -0.06584996, 0.02154825,
        -0.07136868], [
        0.18476107, 0.10267522, -0.05342748, 0.12922333, -0.06481063,
        -0.0133719], [
        -0.20572899, 0.08377647, 0.21218986, 0.00109397, 0.16296447,
        0.03239565]]) if n_points == 4 else (
        make_interesting_starting_points(n_points, tests))

    three: Final[System] = __TO(
        "3oscillators", 6, 1, 2, 2, 1.0,
        tests, training, 5000, 50.0, 5000, 50.0,
        (0, 1, 2, len(training) + len(tests) - 1))
    three.equations = __3_coupled_oscillators  # type: ignore
    return three


#: The 3 oscillators system with 4 training points.
THREE_COUPLED_OSCILLATORS: Final[System] = make_3_couple_oscillators(4)



"""
A game plan assigns teams to games.

A game plan is a two-dimensional matrix `G`. The rows are the time slots.
There is one column for each time. If `G` has value `v` at row `i` and
column `j`, then this means:

- at the time slot `i` ...
- the team with name `j+1` plays
    + no team if `v == 0`,
    + *at home* against the team `v` if `v > 0`, i.e., team `v` travels
      to the home stadium of team `j+1`
    + *away* against the team `-v` if `v < 0`, i.e., team `j+1` travels
      to the home stadium of team `-v` and plays against them there

Indices in matrices are zero-based, i.e., the lowest index for a row `i` is
`0` and the lowest index for a column `j` is also `0`. However, team names
are one-based, i.e., that with `1`. Therefore, we need to translate the
zero-based column index `j` to a team name by adding `1` to it.

This is just a numerical variant of the game plan representation given at
<https://robinxval.ugent.be/RobinX/travelRepo.php>. Indeed, the `str(...)`
representation of a game plan is exactly the table shown there.

Of course, if `G[i, j] = v`, then `G[i, v - 1] = -(j + 1)` should hold if
`v > 0`, for example. Vice versa, if `v < 0` and `G[i, j] = v`, then
`G[i, (-v) - 1] = j + 1` should hold. Such constraints are checked by the
:mod:`~moptipyapps.ttp.errors` objective function.

The corresponding space implementation,
:mod:`~moptipyapps.ttp.game_plan_space`, offers the functionality to convert
strings to game plans as well as to instantiate them in a black-box algorithm.
"""

from io import StringIO
from typing import Final

import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import CSV_SEPARATOR
from pycommons.types import type_error

from moptipyapps.ttp.instance import Instance


class GamePlan(Component, np.ndarray):
    """A game plan, i.e., a solution to the Traveling Tournament Problem."""

    #: the TTP instance
    instance: Instance

    def __new__(cls, instance: Instance) -> "GamePlan":
        """
        Create a solution record for the Traveling Tournament Problem.

        :param cls: the class
        :param instance: the solution record
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)

        n: Final[int] = instance.n_cities  # the number of teams
        # each team plays every other team 'rounds'  times
        n_days: Final[int] = (n - 1) * instance.rounds
        obj: Final[GamePlan] = super().__new__(
            cls, (n_days, n), instance.game_plan_dtype)
        #: the TTP instance
        obj.instance = instance
        return obj

    def __str__(self):
        """
        Convert the game plan to a compact string.

        The first line of the output is a flattened version of this matrix
        with the values being separated by `;`. Then we place an empty line.

        We then put a more easy-to-read representation and follow the pattern
        given at https://robinxval.ugent.be/RobinX/travelRepo.php, which is
        based upon the notation by Easton et al. Here, first, a row with the
        team names separated by spaces is generated. Then, each row contains
        the opponents of these teams, again separated by spaces. If an
        opponent plays at their home, this is denoted by an `@`.
        If a team has no scheduled opponent, then this is denoted as `-`.

        :return: the compact string
        """
        csv: Final[str] = CSV_SEPARATOR
        sep: str = ""
        teams: Final[tuple[str, ...]] = self.instance.teams
        len(teams)

        with StringIO() as sio:
            for k in self.flatten():
                sio.write(sep)
                sio.write(str(k))
                sep = csv

            sio.write("\n\n")

            sep = ""
            for t in teams:
                sio.write(sep)
                sio.write(t)
                sep = " "

            for row in self:
                sio.write("\n")
                sep = ""
                for d in row:
                    sio.write(sep)
                    if d < 0:
                        sio.write(f"@{teams[-d - 1]}")
                    elif d > 0:
                        sio.write(teams[d - 1])
                    else:
                        sio.write("-")
                    sep = " "

            return sio.getvalue()



"""
The Traveling Tournament Problem (TTP).

The Traveling Tournament Problem (TTP) models the logistics of a sports
league, where each game is defined as two teams playing against each other.
In each time slot (let's call that "day") of the tournament, each team has
one game against one other team. One of the two teams of each game will play
"at home," the other "away."

In order to have each of the `n` teams play once against each of the `n-1`
other teams, we need to have `n - 1` "days". So in one round of the
tournament, there are `n - 1` time slots so that each of the teams can play
exactly `n - 1` games, i.e., once against every other team.

If the tournament has two rounds (i.e., is a double round-robin tournament),
then each game appears twice, but with home and away team switched. There are
also constraints for the minimum and maximum number of games at home in a row
and away in a row. There are also constraints for the minimum and maximum
number in between repeating a game (with home/away team switched).

1. David Van Bulck. Minimum Travel Objective Repository. *RobinX: An
   XML-driven Classification for Round-Robin Sports Timetabling.* Faculty of
   Economics and Business Administration at Ghent University, Belgium.
   https://robinxval.ugent.be/. In this repository, you can also find the
   original instance data, useful descriptions, and many links to related
   works.
2. Kelly Easton, George L. Nemhauser, and Michael K. Trick. The Traveling
   Tournament Problem Description and Benchmarks. In *Principles and Practice
   of Constraint Programming (CP'01),*  November 26 - December 1, 2001, Paphos,
   Cyprus, pages 580-584, Berlin/Heidelberg, Germany: Springer.
   ISBN: 978-3-540-42863-3. https://doi.org/10.1007/3-540-45578-7_43.
   https://www.researchgate.net/publication/220270875.
3. Celso C. Ribeiro and Sebastián Urrutia. Heuristics for the Mirrored
   Traveling Tournament Problem. *European Journal of Operational Research*
   (EJOR) 179(3):775-787, June 16, 2007.
   https://doi.org/10.1016/j.ejor.2005.03.061.
   https://optimization-online.org/wp-content/uploads/2004/04/851.pdf.
4. Sebastián Urrutia and Celso C. Ribeiro. Maximizing Breaks and Bounding
   Solutions to the Mirrored Traveling Tournament Problem.
   *Discrete Applied Mathematics* 154(13):1932-1938, August 15, 2006.
   https://doi.org/10.1016/j.dam.2006.03.030.

So far, the following components have been implemented:

1. :mod:`~moptipyapps.ttp.instance` provides the data of a TTP instance,
   including the number of teams, the constraints, and the distance matrix.
   Notice that it is an extension of the Traveling Salesperson Problem
   :mod:`~moptipyapps.tsp.instance` instance data, from which it inherits the
   distance matrix. This data basically describes the starting situation and
   the input data when we try to solve a TTP instance. Also, the module
   provides several of the benchmark instances from
   <https://robinxval.ugent.be/>.
2. :mod:`~moptipyapps.ttp.game_plan` provides a class for holding one
   candidate solution to the TTP, i.e., a game plan. The game plan states, for
   each day and each team, against which other team it will plan (if any). The
   game plan may contain errors, which need to be sorted out by optimization.
3. :mod:`~moptipyapps.ttp.game_plan_space` offers the
   `moptipy` `Space` functionality for such game plans. In other words, it
   allows us to instantiate game plans in a uniform way and to convert them to
   and from strings (which is used when writing log files).
4. :mod:`~moptipyapps.ttp.game_encoding` allows us to decode a permutation
   (potentially with repetitions, see :mod:`~moptipy.spaces.permutations`)
   into a game plan. We therefore can use optimization algorithms and
   operators working on the well-understood space of permutations to produce
   game plans. However, the decoded game plans may have errors, e.g., slots
   without games or violations of the maximum or minimum streak length
   constraints.
5. :mod:`~moptipyapps.ttp.errors` offers an objective function that counts the
   number of such constraint violations in a game plan. If we can minimize it
   to zero, then the game plan is feasible.
"""



"""
An objective function indirectly minimizing the number of bins in packings.

This objective function minimizes the number of bins and maximizes the
"useable" space in the last bin.

Which space is actually useful for our encodings? Let's say we have filled
a bin to a certain degree and somewhere there is a "hole" in the filled area,
but this hole is covered by another object. The area of the hole is not used,
but it also cannot be used anymore. The area that we can definitely use is the
area above the "skyline" of the objects in the bin. The skyline at any
horizontal `x` coordinate be the highest border of any object that intersects
with `x` horizontally. In other words, it is the `y` value at and above which
no other object is located at this `x` coordinate. The area below the skyline
cannot be used anymore. The area above the skyline can.

If we minimize the area below the skyline in the very last bin, then this will
a similar impact as minimizing the overall object area in the last bin (see
:mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_small`). We push
the skyline lower and lower and, if we are lucky, the last bin eventually
becomes empty.

The objective
:mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_lowest_skyline`
works quite similarly to this one, but minimizes the lowest skyline over any
bin.
"""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count_and_last_small import (
    BinCountAndLastSmall,
)
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def bin_count_and_last_skyline(y: np.ndarray, bin_width: int,
                               bin_height: int) -> int:
    """
    Compute the bin count-1 times the bin size + the space below the skyline.

    :param y: the packing
    :param bin_width: the bin width
    :param bin_height: the bin height
    :return: the objective value

    >>> 10*0 + 10*20 + 10*30 + 10*40 + 10*0
    900
    >>> bin_count_and_last_skyline(np.array([[1, 1, 10, 10, 20, 20],
    ...                                      [1, 1, 30, 30, 40, 40],
    ...                                      [1, 1, 20, 20, 30, 30]], int),
    ...                            50, 50)
    900
    >>> 5 * 0 + 5 * 10 + 10 * 20 + 5 * 10 + 25 * 0
    300
    >>> bin_count_and_last_skyline(np.array([[1, 1,  5,  0, 15, 10],
    ...                                      [1, 1, 10, 10, 20, 20],
    ...                                      [1, 1, 15,  0, 25, 10]], int),
    ...                            50, 50)
    300
    >>> 50*50 + 0*10 + 10*20 + 30*0
    2700
    >>> bin_count_and_last_skyline(np.array([[1, 1,  5,  0, 15, 10],
    ...                                      [1, 2, 10, 10, 20, 20],
    ...                                      [1, 1, 15,  0, 25, 10]], int),
    ...                            50, 50)
    2700
    >>> 5 * 0 + 5 * 10 + 3 * 20 + (50 - 13) * 25
    1035
    >>> bin_count_and_last_skyline(np.array([[1, 1,  5,  0, 15, 10],
    ...                                      [1, 1, 10, 10, 20, 20],
    ...                                      [1, 1, 15,  0, 25, 10],
    ...                                      [2, 1, 13, 20, 50, 25]], int),
    ...                            50, 50)
    1035
    >>> 50*50*3 + 25*50
    8750
    >>> bin_count_and_last_skyline(np.array([[1, 1, 0, 0, 10, 10],
    ...                                      [2, 2, 0, 0, 20, 20],
    ...                                      [3, 3, 0, 0, 25, 10],
    ...                                      [4, 4, 0, 0, 50, 25]], int),
    ...                            50, 50)
    8750
    >>> 50*50*3 + 25*10 + 25*0
    7750
    >>> bin_count_and_last_skyline(np.array([[1, 1, 0, 0, 10, 10],
    ...                                      [2, 2, 0, 0, 20, 20],
    ...                                      [3, 4, 0, 0, 25, 10],
    ...                                      [4, 3, 0, 0, 50, 25]], int),
    ...                            50, 50)
    7750
    >>> 50*50*3 + 20*20 + 30*0
    7900
    >>> bin_count_and_last_skyline(np.array([[1, 1, 0, 0, 10, 10],
    ...                                      [2, 4, 0, 0, 20, 20],
    ...                                      [3, 2, 0, 0, 25, 10],
    ...                                      [4, 3, 0, 0, 50, 25]], int),
    ...                            50, 50)
    7900
    >>> 50*50*3 + 20*10 + 30*20 + 20*0
    8300
    >>> bin_count_and_last_skyline(np.array([[1, 1, 0, 0, 10, 10],
    ...                                      [2, 4, 0, 0, 20, 20],
    ...                                      [3, 2, 0, 0, 25, 10],
    ...                                      [2, 4, 10, 20, 30, 30],
    ...                                      [4, 3, 0, 0, 50, 25]], int),
    ...                            50, 50)
    8300
    """
    bins: Final[int] = int(y[:, IDX_BIN].max())
    len_y: Final[int] = len(y)
    bin_size: Final[int] = bin_height * bin_width
    area_under_skyline: int = 0

    cur_left: int = 0
    use_bin: int = max(y[:, IDX_BIN])  # the bin to use
    while cur_left < bin_width:
        use_right = next_left = bin_width
        use_top: int = 0
        for i in range(len_y):
            if y[i, IDX_BIN] != use_bin:
                continue
            left: int = int(y[i, IDX_LEFT_X])
            right: int = int(y[i, IDX_RIGHT_X])
            top: int = int(y[i, IDX_TOP_Y])
            if left <= cur_left < right and top > use_top:
                use_top = top
                use_right = right
            if cur_left < left < next_left:
                next_left = left

        use_right = min(use_right, next_left)
        area_under_skyline += (use_right - cur_left) * use_top
        cur_left = use_right
    return ((bins - 1) * bin_size) + area_under_skyline


class BinCountAndLastSkyline(BinCountAndLastSmall):
    """Compute the number of bins and the useful area in the last bin."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__(instance)
        #: the bin width
        self.__bin_width: Final[int] = instance.bin_width
        #: the bin height
        self.__bin_height: Final[int] = instance.bin_height

    def evaluate(self, x) -> int:
        """
        Evaluate the objective function.

        :param x: the solution
        :return: the bin size and last-bin-small-area factor
        """
        return bin_count_and_last_skyline(
            x, self.__bin_width, self.__bin_height)

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCountAndLowestSkyline`
        :retval "binCountAndLowestSkyline": always
        """
        return "binCountAndLastSkyline"



"""
A Two-Dimensional Bin Packing instance.

This module provides an instance of the two-dimensional bin packing problem
as defined in 2DPackLib [1, 2] as well as the four non-trivial 'Almost Squares
in Almost Squares' instances [6, 7].

All instances of :class:`~moptipyapps.binpacking2d.instance.Instance`
are two-dimensional numpy `ndarrays` with additional attributes.
Each instance has a :attr:`~Instance.name`. Instances also specify a
:attr:`~Instance.bin_width` and :attr:`~Instance.bin_height`.
They define the number :attr:`~Instance.n_different_items` of items with
*different* IDs. Notice that in the 2DPackLib dataset, a benchmark instance
may contain the same item multiple times. Obviously, all items of the same
ID have the exact same width and height, meaning that we only need to store
them once and remember how often they occur. (Notice that the opposite is not
true, i.e., not all items with the same width and height do have the same ID.)
Anyway, the total number :attr:`~Instance.n_items` of items, i.e., the sum of
all the repetitions of all items, is also stored.

The matrix data of the instance class is laid out as follows: There is one row
for each item. The row contains the width of the item, the height of the item,
and the number of times the item will occur. The row at index `i` stands for
the item with ID `i+1`.

Instances can be loaded directly from a 2DPackLib file via
:meth:`Instance.from_2dpacklib`. They can also be loaded from a compact string
representation (via :meth:`Instance.from_compact_str`) and can also be
converted to such a compact representation
(via :meth:`Instance.to_compact_str`). This library ships with a set of
pre-defined instances as resource which can be obtained via
:meth:`Instance.from_resource` and listed via
:meth:`Instance.list_resources`.

We provide the instances of the sets `A` [3], `BENG` [4], and `CLASS` [5]
from 2DPackLib. Additionally, we include the four non-trivial 'Almost Squares
in Almost Squares' instances [6,7].

Initial work on this code has been contributed by Mr. Rui ZHAO (赵睿),
<zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).

1. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele Monaci.
   *2DPackLib*.
   https://site.unibo.it/operations-research/en/research/2dpacklib
2. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele
   Monaci. 2DPackLib: A Two-Dimensional Cutting and Packing Library.
   *Optimization Letters* 16(2):471-480. March 2022.
   https://doi.org/10.1007/s11590-021-01808-y
3. Rita Macedo, Cláudio Alves, and José M. Valério de Carvalho. Arc-Flow
   Model for the Two-Dimensional Guillotine Cutting Stock Problem.
   *Computers & Operations Research* 37(6):991-1001. June 2010.
   https://doi.org/10.1016/j.cor.2009.08.005.
4. Bengt-Erik Bengtsson. Packing Rectangular Pieces - A Heuristic Approach.
   *The Computer Journal* 25(3):353-357, August 1982.
   https://doi.org/10.1093/comjnl/25.3.353
5. J.O. Berkey and P.Y. Wang. Two Dimensional Finite Bin Packing Algorithms.
   *Journal of the Operational Research Society* 38(5):423-429. May 1987.
   https://doi.org/10.1057/jors.1987.70
6. Daan van den Berg, Florian Braam, Mark Moes, Emiel Suilen, and
   Sandjai Bhulai. Almost Squares in Almost Squares: Solving the Final
   Instance. In *The Fifth International Conference on Data Analytics
   (DATA ANALYTICS'16)* October 9-13, 2016, Venice, Italy, pages 69-74.
   IARIA. ISBN: 9781612085104. https://hdl.handle.net/11245/1.545914.
   https://math.vu.nl/~sbhulai/publications/data_analytics2016b.pdf.
7. H. Simonis and B. O'Sullivan. Almost Square Packing in *8th International
   Conference on Integration of AI and OR Techniques in Constraint Programming
   for Combinatorial Optimization Problems* (CPAIOR'11), May 23-27, 2011,
   Berlin, Germany, pages 196-209. Berlin, Germany: Springer-Verlag.
   doi:10.1007/978-3-642-21311-3_19.

>>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
>>> ins.name
'a'
>>> ins.bin_width
100
>>> ins.bin_height
50
>>> ins.dtype
dtype('int8')
>>> ins.n_different_items
3
>>> ins.to_compact_str()
'a;3;100;50;10,5;3,3;5,5'
>>> ins = Instance("b", 100, 50, np.array([[10, 5, 1], [3, 3, 1], [3, 3, 1]]))
>>> ins.name
'b'
>>> ins.bin_width
100
>>> ins.bin_height
50
>>> ins.dtype
dtype('int8')
>>> ins.to_compact_str()
'b;3;100;50;10,5;3,3;3,3'
>>> ins = Instance.from_resource("cl02_020_06")
>>> ins.dtype
dtype('int8')
>>> ins.n_different_items
20
>>> ins.n_items
20
>>> ins = Instance.from_resource("a25")
>>> ins.dtype
dtype('int16')
>>> ins.n_different_items
75
>>> ins.n_items
156
"""

from importlib import resources  # nosem
from os.path import basename
from statistics import quantiles
from typing import Final, Iterable, cast

import moptipy.utils.nputils as npu
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.strings import sanitize_name
from pycommons.io.path import UTF8, Path, file_path
from pycommons.types import check_int_range, check_to_int_range, type_error

#: the instances resource name
INSTANCES_RESOURCE: Final[str] = "instances.txt"

#: the total number of items
N_ITEMS: Final[str] = "nItems"
#: the number of different items.
N_DIFFERENT_ITEMS: Final[str] = "nDifferentItems"
#: the bin width
BIN_WIDTH: Final[str] = "binWidth"
#: the bin height
BIN_HEIGHT: Final[str] = "binHeight"
#: The internal item separator
INTERNAL_SEP: Final[str] = "," if CSV_SEPARATOR == ";" else ";"

#: the index of the width element in an item of an instance
IDX_WIDTH: Final[int] = 0
#: the index of the height element in an item of an instance
IDX_HEIGHT: Final[int] = 1
#: the index of the repetitions element in an item of an instance
IDX_REPETITION: Final[int] = 2

#: the list of instance names of the 2DPackLib bin packing set downloaded
#: from https://site.unibo.it/operations-research/en/research/2dpacklib
#: ('a*','beng*', 'cl*') as well as the four non-trivial 'Almost Squares in
#: Almost Squares' instances ('asqas*').
_INSTANCES: Final[tuple[str, ...]] = (
    "a01", "a02", "a03", "a04", "a05", "a06", "a07", "a08", "a09", "a10",
    "a11", "a12", "a13", "a14", "a15", "a16", "a17", "a18", "a19", "a20",
    "a21", "a22", "a23", "a24", "a25", "a26", "a27", "a28", "a29", "a30",
    "a31", "a32", "a33", "a34", "a35", "a36", "a37", "a38", "a39", "a40",
    "a41", "a42", "a43", "asqas03", "asqas08", "asqas20", "asqas34", "beng01",
    "beng02", "beng03", "beng04", "beng05", "beng06", "beng07", "beng08",
    "beng09", "beng10", "cl01_020_01", "cl01_020_02", "cl01_020_03",
    "cl01_020_04", "cl01_020_05", "cl01_020_06", "cl01_020_07", "cl01_020_08",
    "cl01_020_09", "cl01_020_10", "cl01_040_01", "cl01_040_02", "cl01_040_03",
    "cl01_040_04", "cl01_040_05", "cl01_040_06", "cl01_040_07", "cl01_040_08",
    "cl01_040_09", "cl01_040_10", "cl01_060_01", "cl01_060_02", "cl01_060_03",
    "cl01_060_04", "cl01_060_05", "cl01_060_06", "cl01_060_07", "cl01_060_08",
    "cl01_060_09", "cl01_060_10", "cl01_080_01", "cl01_080_02", "cl01_080_03",
    "cl01_080_04", "cl01_080_05", "cl01_080_06", "cl01_080_07", "cl01_080_08",
    "cl01_080_09", "cl01_080_10", "cl01_100_01", "cl01_100_02", "cl01_100_03",
    "cl01_100_04", "cl01_100_05", "cl01_100_06", "cl01_100_07", "cl01_100_08",
    "cl01_100_09", "cl01_100_10", "cl02_020_01", "cl02_020_02", "cl02_020_03",
    "cl02_020_04", "cl02_020_05", "cl02_020_06", "cl02_020_07", "cl02_020_08",
    "cl02_020_09", "cl02_020_10", "cl02_040_01", "cl02_040_02", "cl02_040_03",
    "cl02_040_04", "cl02_040_05", "cl02_040_06", "cl02_040_07", "cl02_040_08",
    "cl02_040_09", "cl02_040_10", "cl02_060_01", "cl02_060_02", "cl02_060_03",
    "cl02_060_04", "cl02_060_05", "cl02_060_06", "cl02_060_07", "cl02_060_08",
    "cl02_060_09", "cl02_060_10", "cl02_080_01", "cl02_080_02", "cl02_080_03",
    "cl02_080_04", "cl02_080_05", "cl02_080_06", "cl02_080_07", "cl02_080_08",
    "cl02_080_09", "cl02_080_10", "cl02_100_01", "cl02_100_02", "cl02_100_03",
    "cl02_100_04", "cl02_100_05", "cl02_100_06", "cl02_100_07", "cl02_100_08",
    "cl02_100_09", "cl02_100_10", "cl03_020_01", "cl03_020_02", "cl03_020_03",
    "cl03_020_04", "cl03_020_05", "cl03_020_06", "cl03_020_07", "cl03_020_08",
    "cl03_020_09", "cl03_020_10", "cl03_040_01", "cl03_040_02", "cl03_040_03",
    "cl03_040_04", "cl03_040_05", "cl03_040_06", "cl03_040_07", "cl03_040_08",
    "cl03_040_09", "cl03_040_10", "cl03_060_01", "cl03_060_02", "cl03_060_03",
    "cl03_060_04", "cl03_060_05", "cl03_060_06", "cl03_060_07", "cl03_060_08",
    "cl03_060_09", "cl03_060_10", "cl03_080_01", "cl03_080_02", "cl03_080_03",
    "cl03_080_04", "cl03_080_05", "cl03_080_06", "cl03_080_07", "cl03_080_08",
    "cl03_080_09", "cl03_080_10", "cl03_100_01", "cl03_100_02", "cl03_100_03",
    "cl03_100_04", "cl03_100_05", "cl03_100_06", "cl03_100_07", "cl03_100_08",
    "cl03_100_09", "cl03_100_10", "cl04_020_01", "cl04_020_02", "cl04_020_03",
    "cl04_020_04", "cl04_020_05", "cl04_020_06", "cl04_020_07", "cl04_020_08",
    "cl04_020_09", "cl04_020_10", "cl04_040_01", "cl04_040_02", "cl04_040_03",
    "cl04_040_04", "cl04_040_05", "cl04_040_06", "cl04_040_07", "cl04_040_08",
    "cl04_040_09", "cl04_040_10", "cl04_060_01", "cl04_060_02", "cl04_060_03",
    "cl04_060_04", "cl04_060_05", "cl04_060_06", "cl04_060_07", "cl04_060_08",
    "cl04_060_09", "cl04_060_10", "cl04_080_01", "cl04_080_02", "cl04_080_03",
    "cl04_080_04", "cl04_080_05", "cl04_080_06", "cl04_080_07", "cl04_080_08",
    "cl04_080_09", "cl04_080_10", "cl04_100_01", "cl04_100_02", "cl04_100_03",
    "cl04_100_04", "cl04_100_05", "cl04_100_06", "cl04_100_07", "cl04_100_08",
    "cl04_100_09", "cl04_100_10", "cl05_020_01", "cl05_020_02", "cl05_020_03",
    "cl05_020_04", "cl05_020_05", "cl05_020_06", "cl05_020_07", "cl05_020_08",
    "cl05_020_09", "cl05_020_10", "cl05_040_01", "cl05_040_02", "cl05_040_03",
    "cl05_040_04", "cl05_040_05", "cl05_040_06", "cl05_040_07", "cl05_040_08",
    "cl05_040_09", "cl05_040_10", "cl05_060_01", "cl05_060_02", "cl05_060_03",
    "cl05_060_04", "cl05_060_05", "cl05_060_06", "cl05_060_07", "cl05_060_08",
    "cl05_060_09", "cl05_060_10", "cl05_080_01", "cl05_080_02", "cl05_080_03",
    "cl05_080_04", "cl05_080_05", "cl05_080_06", "cl05_080_07", "cl05_080_08",
    "cl05_080_09", "cl05_080_10", "cl05_100_01", "cl05_100_02", "cl05_100_03",
    "cl05_100_04", "cl05_100_05", "cl05_100_06", "cl05_100_07", "cl05_100_08",
    "cl05_100_09", "cl05_100_10", "cl06_020_01", "cl06_020_02", "cl06_020_03",
    "cl06_020_04", "cl06_020_05", "cl06_020_06", "cl06_020_07", "cl06_020_08",
    "cl06_020_09", "cl06_020_10", "cl06_040_01", "cl06_040_02", "cl06_040_03",
    "cl06_040_04", "cl06_040_05", "cl06_040_06", "cl06_040_07", "cl06_040_08",
    "cl06_040_09", "cl06_040_10", "cl06_060_01", "cl06_060_02", "cl06_060_03",
    "cl06_060_04", "cl06_060_05", "cl06_060_06", "cl06_060_07", "cl06_060_08",
    "cl06_060_09", "cl06_060_10", "cl06_080_01", "cl06_080_02", "cl06_080_03",
    "cl06_080_04", "cl06_080_05", "cl06_080_06", "cl06_080_07", "cl06_080_08",
    "cl06_080_09", "cl06_080_10", "cl06_100_01", "cl06_100_02", "cl06_100_03",
    "cl06_100_04", "cl06_100_05", "cl06_100_06", "cl06_100_07", "cl06_100_08",
    "cl06_100_09", "cl06_100_10", "cl07_020_01", "cl07_020_02", "cl07_020_03",
    "cl07_020_04", "cl07_020_05", "cl07_020_06", "cl07_020_07", "cl07_020_08",
    "cl07_020_09", "cl07_020_10", "cl07_040_01", "cl07_040_02", "cl07_040_03",
    "cl07_040_04", "cl07_040_05", "cl07_040_06", "cl07_040_07", "cl07_040_08",
    "cl07_040_09", "cl07_040_10", "cl07_060_01", "cl07_060_02", "cl07_060_03",
    "cl07_060_04", "cl07_060_05", "cl07_060_06", "cl07_060_07", "cl07_060_08",
    "cl07_060_09", "cl07_060_10", "cl07_080_01", "cl07_080_02", "cl07_080_03",
    "cl07_080_04", "cl07_080_05", "cl07_080_06", "cl07_080_07", "cl07_080_08",
    "cl07_080_09", "cl07_080_10", "cl07_100_01", "cl07_100_02", "cl07_100_03",
    "cl07_100_04", "cl07_100_05", "cl07_100_06", "cl07_100_07", "cl07_100_08",
    "cl07_100_09", "cl07_100_10", "cl08_020_01", "cl08_020_02", "cl08_020_03",
    "cl08_020_04", "cl08_020_05", "cl08_020_06", "cl08_020_07", "cl08_020_08",
    "cl08_020_09", "cl08_020_10", "cl08_040_01", "cl08_040_02", "cl08_040_03",
    "cl08_040_04", "cl08_040_05", "cl08_040_06", "cl08_040_07", "cl08_040_08",
    "cl08_040_09", "cl08_040_10", "cl08_060_01", "cl08_060_02", "cl08_060_03",
    "cl08_060_04", "cl08_060_05", "cl08_060_06", "cl08_060_07", "cl08_060_08",
    "cl08_060_09", "cl08_060_10", "cl08_080_01", "cl08_080_02", "cl08_080_03",
    "cl08_080_04", "cl08_080_05", "cl08_080_06", "cl08_080_07", "cl08_080_08",
    "cl08_080_09", "cl08_080_10", "cl08_100_01", "cl08_100_02", "cl08_100_03",
    "cl08_100_04", "cl08_100_05", "cl08_100_06", "cl08_100_07", "cl08_100_08",
    "cl08_100_09", "cl08_100_10", "cl09_020_01", "cl09_020_02", "cl09_020_03",
    "cl09_020_04", "cl09_020_05", "cl09_020_06", "cl09_020_07", "cl09_020_08",
    "cl09_020_09", "cl09_020_10", "cl09_040_01", "cl09_040_02", "cl09_040_03",
    "cl09_040_04", "cl09_040_05", "cl09_040_06", "cl09_040_07", "cl09_040_08",
    "cl09_040_09", "cl09_040_10", "cl09_060_01", "cl09_060_02", "cl09_060_03",
    "cl09_060_04", "cl09_060_05", "cl09_060_06", "cl09_060_07", "cl09_060_08",
    "cl09_060_09", "cl09_060_10", "cl09_080_01", "cl09_080_02", "cl09_080_03",
    "cl09_080_04", "cl09_080_05", "cl09_080_06", "cl09_080_07", "cl09_080_08",
    "cl09_080_09", "cl09_080_10", "cl09_100_01", "cl09_100_02", "cl09_100_03",
    "cl09_100_04", "cl09_100_05", "cl09_100_06", "cl09_100_07", "cl09_100_08",
    "cl09_100_09", "cl09_100_10", "cl10_020_01", "cl10_020_02", "cl10_020_03",
    "cl10_020_04", "cl10_020_05", "cl10_020_06", "cl10_020_07", "cl10_020_08",
    "cl10_020_09", "cl10_020_10", "cl10_040_01", "cl10_040_02", "cl10_040_03",
    "cl10_040_04", "cl10_040_05", "cl10_040_06", "cl10_040_07", "cl10_040_08",
    "cl10_040_09", "cl10_040_10", "cl10_060_01", "cl10_060_02", "cl10_060_03",
    "cl10_060_04", "cl10_060_05", "cl10_060_06", "cl10_060_07", "cl10_060_08",
    "cl10_060_09", "cl10_060_10", "cl10_080_01", "cl10_080_02", "cl10_080_03",
    "cl10_080_04", "cl10_080_05", "cl10_080_06", "cl10_080_07", "cl10_080_08",
    "cl10_080_09", "cl10_080_10", "cl10_100_01", "cl10_100_02", "cl10_100_03",
    "cl10_100_04", "cl10_100_05", "cl10_100_06", "cl10_100_07", "cl10_100_08",
    "cl10_100_09", "cl10_100_10")


def __divide_based_on_size(instances: Iterable[str],
                           divisions: int) -> list[list[str]]:
    """
    Divide the instances based on their size.

    :param instances: the instances
    :param divisions: the number of divisions
    :return: the divided instances
    """
    insts: list[str] = list(instances)
    if len(insts) <= 0:
        return [insts]

    try:
        loaded: list[Instance] = [Instance.from_resource(n) for n in insts]
    except (OSError, ValueError):
        return [insts]

    # get the quantiles = group divisions
    qants: list[float | int] = quantiles((
        inst.n_items for inst in loaded), n=divisions, method="inclusive")

    # now put the instances into the groups
    groups: list[list[str]] = []
    for inst in loaded:
        inst_size = inst.n_items
        idx: int = 0
        for q in qants:
            if q > inst_size:
                break
            idx += 1
        while idx >= len(groups):
            groups.append([])
        groups[idx].append(str(inst))

    # remove useless groups
    for idx in range(len(groups) - 1, -1, -1):
        if len(groups[idx]) <= 0:
            del groups[idx]
    return groups


def _make_instance_groups(instances: Iterable[str]) \
        -> tuple[tuple[str, str | None, tuple[str, ...]], ...]:
    """
    Make the standard instance groups from an instance name list.

    :return: the instance groups
    """
    groups: list[tuple[str, str | None, tuple[str, ...]]] = []

    a_divided: list[list[str]] = __divide_based_on_size(sorted(
        b for b in instances if b.startswith("a")
        and (len(b) == 3) and b[-1].isdigit()
        and b[-2].isdigit()), 3)
    if len(a_divided) > 0:
        subnames: list[str | None] = [None] if (len(a_divided) <= 1) else (
            ["small", "large"] if (len(a_divided) <= 2) else
            ["small", "med", "large"])
        for a_idx, a_group in enumerate(a_divided):
            v: tuple[str, str | None, tuple[str, ...]] = (
                "a", subnames[a_idx], tuple(a_group))
            if len(v[2]) > 0:
                groups.append(v)

    v = ("beng", "1-8", tuple(sorted(
        b for b in instances if b.startswith("beng")
        and (int(b[-2:]) < 9))))
    if len(v[2]) > 0:
        groups.append(v)

    v = ("beng", "9-10", tuple(sorted(
        b for b in instances if b.startswith("beng")
        and (int(b[-2:]) >= 9))))
    if len(v[2]) > 0:
        groups.append(v)

    for i in range(1, 11):
        name: str = f"class {i}"
        preprefix: str = f"cl0{i}" if i < 10 else f"cl{i}"
        for n in (20, 40, 60, 80, 100):
            prefix: str = f"{preprefix}_0{n}_" \
                if n < 100 else f"{preprefix}_{n}_"
            v = (name, str(n), tuple(sorted(
                b for b in _INSTANCES if b.startswith(prefix))))
            if len(v[2]) > 0:
                groups.append(v)

    v = ("asqas", None, tuple(sorted(
        b for b in _INSTANCES if b.startswith("asqas"))))
    if len(v[2]) > 0:
        groups.append(v)

    all_set: set[str] = set()
    for g in groups:
        all_set.update(g[2])
    inst_set: set[str] = set(instances)
    if all_set != inst_set:
        raise ValueError(f"group instances is {all_set!r} but "
                         f"instance set is {inst_set!r}!")
    return tuple(groups)


def __cutsq(matrix: np.ndarray) -> list[int]:
    """
    Cut all items into squares via the CUTSQ procedure.

    :param matrix: the item matrix
    :return: the list of squares

    >>> __cutsq(np.array([[14, 12, 1]], int))
    [12, 2, 2, 2, 2, 2, 2]

    >>> __cutsq(np.array([[14, 12, 2]], int))
    [12, 12, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    """
    # create list of items in horizontal orientation
    j_sq: Final[list[int]] = []  # the list of squares (width = height)
    s: Final[list[int]] = []  # a temporary list
    for row in matrix:
        w: int = int(row[0])
        h: int = int(row[1])
        if h > w:
            w, h = h, w
        while h > 1:
            k: int = w // h
            for _ in range(k):
                s.append(h)
            w, h = h, w - (k * h)
        times: int = int(row[2])
        j_sq.extend(s * times if times > 1 else s)
        s.clear()

    j_sq.sort(reverse=True)  # sort the squares in decreasing size
    return j_sq


def __lb_q(bin_width: int, bin_height: int, q: int, j_js: list[int]) -> int:
    """
    Compute the lower bound for a given q.

    :param bin_width: the bin width
    :param bin_height: the bin height
    :param q: the parameter q
    :param j_js: the sorted square list
    :return: the lower bound

    >>> jj = [18, 18, 12, 12, 11, 11, 11, 11, 11, 7, 7, 7, 7, 7, 7]
    >>> len(jj)
    15
    >>> __lb_q(23, 20, 6, jj)
    6
    """
    m: Final[int] = len(j_js)
    half_width: Final[float] = bin_width / 2
    half_height: Final[float] = bin_height / 2
    width_m_q: Final[int] = bin_width - q

    # First we compute sets S1 to S4.
    s1: list[int] = []  # S1 from Equation 2
    s2: list[int] = []  # S2 from Equation 3
    s3: list[int] = []  # S2 from Equation 4
    s4: list[int] = []  # S2 from Equation 5

    for i in range(m):
        l_i: int = j_js[i]
        if l_i > width_m_q:
            s1.append(i)  # Equation 2
        elif l_i > half_width:
            s2.append(i)  # Equation 3
        elif l_i > half_height:
            s3.append(i)  # Equation 4
        elif l_i >= q:
            s4.append(i)  # Equation 5
        else:
            break

    # compute set S23 as in Theorem 3 under Equation 7
    height_m_q: Final[int] = bin_height - q
    s23: Final[list[int]] = [j for j in (s2 + s3) if j_js[j] > height_m_q]

    # Now we sort S2 by non-increasing value of residual space.
    s2.reverse()  # = .sort(key=lambda i: bin_width - j_js[i], reverse=True)

    # Now we compute S3 - ^S3^
    s3_minus_s3d: list[int] = s3.copy()
    for i in s2:
        residual: int = bin_width - j_js[i]
        not_found: bool = True
        for j, idx in enumerate(s3_minus_s3d):
            needs: int = j_js[idx]
            if needs <= residual:
                del s3_minus_s3d[j]
                not_found = False
                break
        if not_found:
            break

    sum_s3_l: int = sum(j_js[i] for i in s3_minus_s3d)
    b1 = sum_s3_l // bin_width
    if (b1 * bin_width) < sum_s3_l:
        b1 = b1 + 1

    len_s3: int = len(s3_minus_s3d)
    div: int = bin_width // ((bin_height // 2) + 1)
    b2 = len_s3 // div
    if (b2 * div) < len_s3:
        b2 = b2 + 1

    l_tilde: Final[int] = len(s2) + (b1 if b1 >= b2 else b2)  # Equation 6.
    bound: int = len(s1) + l_tilde

    # Now compute the final bound based on Theorem 3 / Equation 7.
    bin_size: Final[int] = bin_width * bin_height
    denom: int = sum(j_js[i] ** 2 for i in (s2 + s3 + s4)) \
        - ((bin_size * l_tilde) - sum(j_js[i] * (
            bin_height - j_js[i]) for i in s23))
    if denom > 0:
        b = denom // bin_size
        if (b * bin_size) < denom:
            b = b + 1
        bound = bound + b

    return bound


def _lower_bound_damv(bin_width: int, bin_height: int,
                      matrix: np.ndarray) -> int:
    """
    Compute the lower bound as defined by Dell'Amico et al.

    :param bin_width: the bin width
    :param bin_height: the bin height
    :param matrix: the item matrix
    :return: the lower bound

    >>> mat = np.array([[10, 5, 1], [3, 3, 1], [3, 3, 1]])
    >>> _lower_bound_damv(23, 20, mat)
    1

    >>> mat = np.array([[20, 5, 3], [13, 23, 1], [13, 9, 3]])
    >>> _lower_bound_damv(23, 20, mat)
    3
    """
    # ensure horizontal orientation (width >= height)
    if bin_height > bin_width:
        bin_width, bin_height = bin_height, bin_width
    j_sq: Final[list[int]] = __cutsq(matrix)
    res: Final[int] = max(__lb_q(bin_width, bin_height, q, j_sq)
                          for q in range((bin_height // 2) + 1))
    return max(1, res)


class Instance(Component, np.ndarray):
    """
    An instance of the 2D Bin Packing Problem.

    Each row of the matrix contains three values: 1. the item's width, 2. the
    item's height, 3. how often the item occurs.
    """

    #: the name of the instance
    name: str
    #: the total number of items (including repetitions)
    n_items: int
    #: the number of different items
    n_different_items: int
    #: the total area occupied by all items
    total_item_area: int
    #: the bin width
    bin_width: int
    #: the bin height
    bin_height: int
    #: the minimum number of bins that this instance requires
    lower_bound_bins: int

    def __new__(cls, name: str,
                bin_width: int, bin_height: int,
                matrix: np.ndarray | list[list[int]]) -> "Instance":
        """
        Create an instance of the 2D bin packing problem.

        :param cls: the class
        :param name: the name of the instance
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param matrix: the matrix with the data (will be copied); rows have
            the form `(width, height, repetitions)`
        """
        use_name: Final[str] = sanitize_name(name)
        if name != use_name:
            raise ValueError(f"Name {name!r} is not a valid name.")

        check_int_range(bin_width, "bin_width", 1, 1_000_000_000_000)
        check_int_range(bin_height, "bin_height", 1, 1_000_000_000_000)
        max_dim: Final[int] = max(bin_width, bin_height)
        min_dim: Final[int] = min(bin_width, bin_height)

        n_different_items: Final[int] = check_int_range(
            len(matrix), "n_different_items", 1, 100_000_000)
        use_shape: Final[tuple[int, int]] = (n_different_items, 3)

        if isinstance(matrix, np.ndarray):
            if not npu.is_np_int(matrix.dtype):
                raise ValueError(
                    "Matrix must have an integer type, but is of type "
                    f"{str(matrix.dtype)!r} in instance {name!r}.")
            if matrix.shape != use_shape:
                raise ValueError(
                    f"Invalid shape {str(matrix.shape)!r} of matrix: "
                    "must have three columns and two dimensions, must be "
                    f"equal to {use_shape} in instance {name!r}.")
        elif not isinstance(matrix, list):
            raise type_error(matrix, "matrix", np.ndarray)

        n_items: int = 0
        max_size: int = -1
        item_area: int = 0
        for i in range(n_different_items):
            row = matrix[i]
            if not isinstance(row, list | np.ndarray):
                raise type_error(
                    row, f"{row} at index {i} in {use_name!r}",
                    (list, np.ndarray))
            if len(row) != 3:
                raise ValueError(
                    f"invalid row {row} at index {i} in {use_name!r}.")
            width, height, repetitions = row
            width = check_int_range(int(width), "width", 1, max_dim)
            height = check_int_range(int(height), "height", 1, max_dim)
            repetitions = check_int_range(int(repetitions), "repetitions",
                                          1, 100_000_000)
            item_area += (width * height * repetitions)
            max_size = max(width, height)
            if (width > min_dim) and (height > min_dim):
                raise ValueError(
                    f"object with width={width} and height={height} does "
                    f"not fit into bin with width={width} and "
                    f"height={height}.")
            n_items += repetitions

        obj: Final[Instance] = super().__new__(
            cls, use_shape, int_range_to_dtype(
                min_value=0, max_value=max(max_dim + max_size, n_items),
                force_signed=True))
        for i in range(n_different_items):
            obj[i, :] = matrix[i]

        #: the name of the instance
        obj.name = use_name
        #: the number of different items
        obj.n_different_items = n_different_items
        #: the total number of items, i.e., the number of different items
        #: multiplied with their repetition counts
        obj.n_items = check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000)
        #: the height of the bins
        obj.bin_height = bin_height
        #: the width of the bins
        obj.bin_width = bin_width
        #: the total area occupied by all items
        obj.total_item_area = item_area

# We need at least as many bins such that their area is big enough
# for the total area of the items.
        bin_area: int = bin_height * bin_width
        lower_bound_geo: int = item_area // bin_area
        if (lower_bound_geo * bin_area) < item_area:
            lower_bound_geo += 1
        lower_bound_geo = check_int_range(
            lower_bound_geo, "lower_bound_bins_geometric",
            1, 1_000_000_000_000)

# We now compute the lower bound by Dell'Amico et al.
        lower_bound_damv = check_int_range(_lower_bound_damv(
            bin_width, bin_height, obj), "lower_bound_bins_damv",
            1, 1_000_000_000_000)

# The overall computed lower bound is the maximum of the geometric and the
# Dell'Amico lower bound.
        obj.lower_bound_bins = max(lower_bound_damv, lower_bound_geo)
        return obj

    def __str__(self):
        """
        Get the name of this instance.

        :return: the name of this instance
        """
        return self.name

    def get_standard_item_sequence(self) -> list[int]:
        """
        Get the standardized item sequence.

        :return: the standardized item sequence

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.get_standard_item_sequence()
        [1, 2, 3]
        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 2]])
        >>> ins.get_standard_item_sequence()
        [1, 2, 3, 3]
        >>> ins = Instance("a", 100, 50, [[10, 5, 2], [3, 3, 3], [5, 5, 4]])
        >>> ins.get_standard_item_sequence()
        [1, 1, 2, 2, 2, 3, 3, 3, 3]
        """
        base_string: Final[list[int]] = []
        for i in range(1, self.n_different_items + 1):
            for _ in range(self[i - 1, IDX_REPETITION]):
                base_string.append(int(i))
        return base_string

    @staticmethod
    def from_2dpacklib(file: str) -> "Instance":
        """
        Load a problem instance from the 2dpacklib format.

        :param file: the file path
        :return: the instance
        """
        path: Final[Path] = file_path(file)
        name: str = basename(path).lower()
        if name.endswith(".ins2d"):
            name = sanitize_name(name[:-6])

        lines: Final[list[str]] = str.splitlines(path.read_all_str())
        n_different_items: Final[int] = check_to_int_range(
            lines[0], "n_different_items", 1, 100_000_000)

        wh: Final[str] = lines[1]
        spaceidx: Final[int] = wh.index(" ")
        if spaceidx <= 0:
            raise ValueError("Did not find space in second line.")
        bin_width: Final[int] = check_to_int_range(
            wh[:spaceidx], "bin_width", 1, 1_000_000_000_000)
        bin_height: Final[int] = check_to_int_range(
            wh[spaceidx + 1:], "bin_height", 1, 1_000_000_000_000)
        del lines[0:2]

        max_dim: Final[int] = max(bin_width, bin_height)
        data: Final[list[list[int]]] = []
        old_item_id: int = 0
        for line in lines:
            text: list[str] = line.split(" ")
            itemid: int = check_to_int_range(
                text[0], "item-id", 1, n_different_items)
            if itemid != (old_item_id + 1):
                raise ValueError(
                    f"non-sequential item id {itemid} after {old_item_id}!")
            old_item_id = itemid
            width: int = check_to_int_range(text[1], "width", 1, max_dim)
            height: int = check_to_int_range(text[2], "height", 1, max_dim)
            count: int = 1 if len(text) <= 3 else \
                check_to_int_range(text[3], "count", 1, 1_000_000)
            data.append([width, height, count])
        data.sort()
        return Instance(name, bin_width, bin_height, data)

    def to_compact_str(self) -> str:
        """
        Convert the instance to a compact string.

        The format of the string is a single line of semi-colon separated
        values. The values are: `name;n_items;bin_width;bin_height`,
        followed by the sequence of items, each in the form of
        `;width,heigh[,times]`, where `times` is optional and only added
        if the item occurs more than once.

        :return: a single line string with all the instance data

        >>> ins = Instance("x", 500, 50, [[3, 5, 1], [2, 5, 2]])
        >>> ins.to_compact_str()
        'x;2;500;50;3,5;2,5,2'
        >>> ins.n_different_items
        2
        """
        lst: Final[list[str]] = [self.name, str(self.n_different_items),
                                 str(self.bin_width), str(self.bin_height)]
        for i in range(self.n_different_items):
            width: int = self[i, IDX_WIDTH]
            height: int = self[i, IDX_HEIGHT]
            repetitions: int = self[i, IDX_REPETITION]
            lst.append(
                f"{width}{INTERNAL_SEP}{height}" if repetitions == 1 else
                f"{width}{INTERNAL_SEP}{height}{INTERNAL_SEP}{repetitions}")
        return CSV_SEPARATOR.join(lst)

    @staticmethod
    def from_compact_str(data: str) -> "Instance":
        """
        Transform a compact string back to an instance.

        :param data: the string data
        :return: the instance

        >>> ins = Instance("x", 500, 50, [[3, 5, 1], [2, 5, 2]])
        >>> Instance.from_compact_str(ins.to_compact_str()).to_compact_str()
        'x;2;500;50;3,5;2,5,2'
        """
        if not isinstance(data, str):
            raise type_error(data, "data", str)
        text: Final[list[str]] = data.split(CSV_SEPARATOR)
        name: Final[str] = text[0]
        n_different_items: Final[int] = check_to_int_range(
            text[1], "n_different_items", 1, 100_000_000)
        bin_width: Final[int] = check_to_int_range(
            text[2], "bin_width", 1, 1_000_000_000_000)
        bin_height: Final[int] = check_to_int_range(
            text[3], "bin_height", 1, 1_000_000_000_000)
        max_dim: Final[int] = max(bin_width, bin_height)
        items: list[list[int]] = []
        for i in range(4, n_different_items + 4):
            s: list[str] = text[i].split(INTERNAL_SEP)
            row: list[int] = [
                check_to_int_range(s[IDX_WIDTH], "width", 1, max_dim),
                check_to_int_range(s[IDX_HEIGHT], "height", 1, max_dim),
                1 if len(s) <= IDX_REPETITION else
                check_to_int_range(
                    s[IDX_REPETITION], "times", 1, 100_000_000)]
            items.append(row)
        return Instance(name, bin_width, bin_height, items)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters describing this bin packing instance to the logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("I") as kv:
        ...         Instance.from_resource("beng05").log_parameters_to(kv)
        ...     print(repr('@'.join(l.get_log())))
        'BEGIN_I@name: beng05@class: moptipyapps.binpacking2d\
.instance.Instance@nItems: 100@nDifferentItems: 100@binWidth: 25\
@binHeight: 10@dtype: b@END_I'
        """
        super().log_parameters_to(logger)
        logger.key_value(N_ITEMS, self.n_items)
        logger.key_value(N_DIFFERENT_ITEMS, self.n_different_items)
        logger.key_value(BIN_WIDTH, self.bin_width)
        logger.key_value(BIN_HEIGHT, self.bin_height)
        logger.key_value(npu.KEY_NUMPY_TYPE, self.dtype.char)

    @staticmethod
    def list_resources() -> tuple[str, ...]:
        """
        Get a tuple of all the instances available as resource.

        :return: the tuple with the instance names

        >>> len(list(Instance.list_resources()))
        557
        """
        return _INSTANCES

    @staticmethod
    def list_resources_groups() -> tuple[tuple[
            str, str | None, tuple[str, ...]], ...]:
        """
        List the instance groups in the resources.

        One problem of the benchmark set for 2-dimensional bin packing is that
        it has many instances:

        >>> len(Instance.list_resources())
        557

        With this function, we can group several of these instances together
        in a way that is compliant with literature in order to then compute
        statistics over these groups. Presenting data gathered over...

        >>> len(Instance.list_resources_groups())
        56

        ...groups is much easier than dealing with over 500 instances.

        :return: the instance groups, in a two level hierarchy. The result is
            a sequence of tuples. Each tuple has the top-level group name and,
            optionally, a second-level group name (or `None` if no second
            level group exists). The third tuple element is a sequence of
            instance names.

        >>> [(v[0], v[1], len(v[2])) for v in
        ...     Instance.list_resources_groups()]
        [('a', 'small', 14), ('a', 'med', 14), ('a', 'large', 15), \
('beng', '1-8', 8), ('beng', '9-10', 2), ('class 1', '20', 10), \
('class 1', '40', 10), ('class 1', '60', 10), ('class 1', '80', 10), \
('class 1', '100', 10), ('class 2', '20', 10), ('class 2', '40', 10), \
('class 2', '60', 10), ('class 2', '80', 10), ('class 2', '100', 10), \
('class 3', '20', 10), ('class 3', '40', 10), ('class 3', '60', 10), \
('class 3', '80', 10), ('class 3', '100', 10), ('class 4', '20', 10), \
('class 4', '40', 10), ('class 4', '60', 10), ('class 4', '80', 10), \
('class 4', '100', 10), ('class 5', '20', 10), ('class 5', '40', 10), \
('class 5', '60', 10), ('class 5', '80', 10), ('class 5', '100', 10), \
('class 6', '20', 10), ('class 6', '40', 10), ('class 6', '60', 10), \
('class 6', '80', 10), ('class 6', '100', 10), ('class 7', '20', 10), \
('class 7', '40', 10), ('class 7', '60', 10), ('class 7', '80', 10), \
('class 7', '100', 10), ('class 8', '20', 10), ('class 8', '40', 10), \
('class 8', '60', 10), ('class 8', '80', 10), ('class 8', '100', 10), \
('class 9', '20', 10), ('class 9', '40', 10), ('class 9', '60', 10), \
('class 9', '80', 10), ('class 9', '100', 10), ('class 10', '20', 10), \
('class 10', '40', 10), ('class 10', '60', 10), ('class 10', '80', 10), \
('class 10', '100', 10), ('asqas', None, 4)]
        """
        obj: object = Instance.list_resources_groups
        attr: str = "__gs"
        if not hasattr(obj, attr):
            gs: tuple[tuple[str, str | None, tuple[str, ...]], ...] =\
                _make_instance_groups(Instance.list_resources())
            setattr(obj, attr, gs)
            return gs
        return cast(tuple[tuple[str, str | None, tuple[str, ...]], ...],
                    getattr(obj, attr))

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load an instance from a resource.

        :param name: the name string
        :return: the instance

        >>> Instance.from_resource("a01").to_compact_str()
        'a01;2;2750;1220;463,386,18;1680,420,6'
        >>> Instance.from_resource("a07").to_compact_str()
        'a07;5;2750;1220;706,286,8;706,516,16;986,433,10;1120,486,\
12;1120,986,12'
        >>> Instance.from_resource("a08") is Instance.from_resource("a08")
        True
        >>> Instance.from_resource("a08") is Instance.from_resource("a09")
        False
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        container: Final = Instance.from_resource
        inst_attr: Final[str] = f"__inst_{name}"
        if hasattr(container, inst_attr):  # instance loaded?
            return cast(Instance, getattr(container, inst_attr))
        text_attr: Final[str] = f"__text_{INSTANCES_RESOURCE}"
        text: list[str]
        total_attr: Final[str] = "__total_insts"
        if hasattr(container, text_attr):  # ok, we got the text already
            text = cast(list[str], getattr(container, text_attr))
        else:  # the first time we load the text
            with resources.files(__package__).joinpath(
                    INSTANCES_RESOURCE).open("r", encoding=UTF8) as stream:
                text = [line for line in (line_raw.strip() for line_raw
                                          in stream.readlines())
                        if len(line) > 0]
            setattr(container, text_attr, text)
            setattr(container, total_attr, 0)  # so far, no instance

        imax: int = len(text)
        imin: int = 0
        while imin <= imax:  # binary search for instance
            imid: int = (imax + imin) // 2
            line: str = text[imid]
            idx: int = line.index(CSV_SEPARATOR)
            prefix: str = line[0:idx]
            if prefix == name:
                instance = Instance.from_compact_str(line)
                setattr(container, inst_attr, instance)
                got: int = getattr(container, total_attr)
                got = got + 1
                if got >= len(_INSTANCES):  # got all instances, can del text
                    delattr(container, total_attr)
                    delattr(container, text_attr)
                return instance
            if prefix < name:
                imin = imid + 1
            else:
                imax = imid - 1
        raise ValueError(f"instance {name!r} not found.")



"""Plot a packing into one figure."""
from collections import Counter
from typing import Callable, Final, Iterable

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore
from matplotlib.patches import Rectangle  # type: ignore
from matplotlib.text import Text  # type: ignore
from pycommons.types import type_error

from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_ID,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
    Packing,
)


def default_packing_item_str(item_id: int, item_index: int,
                             item_in_bin_index: int) -> Iterable[str]:
    """
    Get a packing item string(s).

    The default idea is to include the item id, the index of the item in the
    bin, and the overall index of the item. If the space is insufficient,
    we remove the latter or the latter two. Hence, this function returns a
    tuple of three strings.

    :param item_id: the ID of the packing item
    :param item_index: the item index
    :param item_in_bin_index: the index of the item in its bin
    :return: the string
    """
    return (f"{item_id}/{item_in_bin_index}/{item_index}",
            f"{item_id}/{item_in_bin_index}", str(item_id))


def plot_packing(packing: Packing | str,
                 max_rows: int = 3,
                 max_bins_per_row: int = 3,
                 default_width_per_bin: float | int | None = 8.6,
                 max_width: float | int | None = 8.6,
                 default_height_per_bin: float | int | None =
                 5.315092303249095,
                 max_height: float | int | None = 9,
                 packing_item_str: Callable[
                     [int, int, int], str | Iterable[str]] =
                 default_packing_item_str,
                 importance_to_font_size_func: Callable[[int], float] =
                 pd.importance_to_font_size,
                 dpi: float | int | None = 384.0) -> Figure:
    """
    Plot a packing.

    Each item is drawn in a different color. Each item rectangle includes, if
    there is enough space, the item-ID. If there is more space, also the index
    of the item inside the bin (starting at 1) is included. If there is yet
    more space, even the overall index of the item is included.

    :param packing: the packing or the file to load it from
    :param max_rows: the maximum number of rows
    :param max_bins_per_row: the maximum number of bins per row
    :param default_width_per_bin: the optional default width of a column
    :param max_height: the maximum height
    :param default_height_per_bin: the optional default height per row
    :param max_width: the maximum width
    :param packing_item_str: the function converting an item id,
        item index, and item-in-bin index to a string or sequence of strings
        (of decreasing length)
    :param importance_to_font_size_func: the function converting
        importance values to font sizes
    :param dpi: the dpi value
    :returns: the Figure object to allow you to add further plot elements
    """
    if isinstance(packing, str):
        packing = Packing.from_log(packing)
    if not isinstance(packing, Packing):
        raise type_error(packing, "packing", (Packing, str))
    if not callable(packing_item_str):
        raise type_error(packing_item_str, "packing_item_str", call=True)

    # allocate the figure ... this is hacky for now
    figure, bin_figures = pu.create_figure_with_subplots(
        items=packing.n_bins, max_items_per_plot=1, max_rows=max_rows,
        max_cols=max_bins_per_row, min_rows=1, min_cols=1,
        default_width_per_col=default_width_per_bin, max_width=max_width,
        default_height_per_row=default_height_per_bin, max_height=max_height,
        dpi=dpi)

    # initialize the different plots
    bin_width: Final[int] = packing.instance.bin_width
    bin_height: Final[int] = packing.instance.bin_height
    axes_list: Final[list[Axes]] = []
    for the_axes, _, _, _, _, _ in bin_figures:
        axes = pu.get_axes(the_axes)
        axes_list.append(axes)
        axes.set_ylim(0, bin_width)  # pylint: disable=E1101
        axes.set_ybound(0, bin_height)  # pylint: disable=E1101
        axes.set_xlim(0, bin_width)  # pylint: disable=E1101
        axes.set_xbound(0, bin_width)  # pylint: disable=E1101
        axes.set_aspect("equal", None, "C")  # pylint: disable=E1101
        axes.tick_params(  # pylint: disable=E1101
            left=False, bottom=False, labelleft=False, labelbottom=False)

    # get the color and font styles
    colors: Final[tuple] = pd.distinct_colors(
        packing.instance.n_different_items)
    font_size: Final[float] = importance_to_font_size_func(-1)

    # get the transforms needed to obtain text dimensions
    renderers: Final[list] = [pu.get_renderer(axes) for axes in axes_list]
    inverse: Final[list] = [
        axes.transData.inverted()  # type: ignore  # pylint: disable=E1101
        for axes in axes_list]

    z_order: int = 0  # the z-order of all drawing elements

    # we now plot the items one-by-one
    bin_counters: Counter[int] = Counter()
    for item_index in range(packing.instance.n_items):
        item_id: int = packing[item_index, IDX_ID]
        item_bin: int = packing[item_index, IDX_BIN]
        x_left: int = packing[item_index, IDX_LEFT_X]
        y_bottom: int = packing[item_index, IDX_BOTTOM_Y]
        x_right: int = packing[item_index, IDX_RIGHT_X]
        y_top: int = packing[item_index, IDX_TOP_Y]
        item_in_bin_index: int = bin_counters[item_bin] + 1
        bin_counters[item_bin] = item_in_bin_index
        width: int = x_right - x_left
        height: int = y_top - y_bottom

        background = colors[item_id - 1]
        foreground = pd.text_color_for_background(colors[item_id - 1])

        axes = axes_list[item_bin - 1]
        rend = renderers[item_bin - 1]
        inv = inverse[item_bin - 1]

        axes.add_artist(Rectangle(  # paint the item's rectangle
            xy=(x_left, y_bottom), width=width, height=height,
            facecolor=background, linewidth=0.75, zorder=z_order,
            edgecolor="black"))
        z_order += 1
        x_center: float = 0.5 * (x_left + x_right)
        y_center: float = 0.5 * (y_bottom + y_top)

# get the box label string or string sequence
        strs = packing_item_str(item_id, item_index + 1, item_in_bin_index)
        if isinstance(strs, str):
            strs = [strs]
        elif not isinstance(strs, Iterable):
            raise type_error(
                strs, f"packing_item_str({item_id}, {item_index}, "
                      f"{item_in_bin_index})", (str, Iterable))

# iterate over the possible box label strings
        for i, item_str in enumerate(strs):
            if not isinstance(item_str, str):
                raise type_error(
                    str, f"packing_item_str({item_id}, {item_index}, "
                         f"{item_in_bin_index})[{i}]", str)
# Get the size of the text using a temporary text  that gets immediately
# deleted again.
            tmp: Text = axes.text(x=x_center, y=y_center, s=item_str,
                                  fontsize=font_size,
                                  color=foreground,
                                  horizontalalignment="center",
                                  verticalalignment="baseline")
            bb_bl = inv.transform_bbox(tmp.get_window_extent(
                renderer=rend))
            Artist.set_visible(tmp, False)
            Artist.remove(tmp)
            del tmp

# Check if this text did fit into the rectangle.
            if (bb_bl.width < (0.97 * width)) and \
                    (bb_bl.height < (0.97 * height)):
                # OK, there is enough space. Let's re-compute the y offset
                # to do proper alignment using another temporary text.
                tmp = axes.text(x=x_center, y=y_center, s=item_str,
                                fontsize=font_size,
                                color=foreground,
                                horizontalalignment="center",
                                verticalalignment="bottom")
                bb_bt = inv.transform_bbox(tmp.get_window_extent(
                    renderer=rend))
                Artist.set_visible(tmp, False)
                Artist.remove(tmp)
                del tmp

# Now we can really print the actual text with a more or less  nice vertical
# alignment.
                adj = bb_bl.y0 - bb_bt.y0
                if adj < 0:
                    y_center += adj / 3

                axes.text(x=x_center, y=y_center, s=item_str,
                          fontsize=font_size,
                          color=foreground,
                          horizontalalignment="center",
                          verticalalignment="center",
                          zorder=z_order)
                z_order += 1
                break  # We found a text that fits, so we can quit.
    return figure



"""Perform tests on the Two-Dimensional Bin Packing Problem."""

from typing import Callable, Final, Iterable, cast

import numpy as np
import numpy.random as rnd
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding
from moptipy.api.objective import Objective
from moptipy.operators.signed_permutations.op0_shuffle_and_flip import (
    Op0ShuffleAndFlip,
)
from moptipy.spaces.signed_permutations import SignedPermutations
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.objective import validate_objective
from moptipy.tests.space import validate_space
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipyapps.binpacking2d.encodings.ibl_encoding_1 import (
    ImprovedBottomLeftEncoding1,
)
from moptipyapps.binpacking2d.encodings.ibl_encoding_2 import (
    ImprovedBottomLeftEncoding2,
)
from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count_and_last_empty import (
    BinCountAndLastEmpty,
)
from moptipyapps.binpacking2d.packing import Packing
from moptipyapps.binpacking2d.packing_space import PackingSpace

#: the internal random number generator
__RANDOM: Final[Generator] = default_rng()


def binpacking_instances_for_tests(
        random: Generator = __RANDOM) -> Iterable[str]:
    """
    Get a sequence of 2D Bin Packing instances to test on.

    :param random: the random number generator to use
    :returns: an iterable of 2D Bin Packing instance names
    """
    ri = random.integers
    insts: set[str] = {
        "a04", "a08", "beng10", f"a0{ri(1, 10)}", f"a1{ri(1, 10)}",
        f"a2{ri(1, 10)}", f"a3{ri(1, 10)}", f"a4{ri(1, 4)}",
        f"beng0{ri(1, 10)}", f"cl01_080_0{ri(1, 10)}",
        f"cl06_020_0{ri(1, 10)}", f"cl09_040_0{ri(1, 10)}",
        f"cl10_100_0{ri(1, 10)}"}
    insn: list[str] = list(Instance.list_resources())
    while len(insts) < 16:
        insts.add(insn.pop(ri(len(insn))))
    use_insts: list[str] = list(insts)
    random.shuffle(cast(list, use_insts))
    return use_insts


def make_packing_valid(inst: Instance,
                       random: Generator = __RANDOM) \
        -> Callable[[Generator, Packing], Packing]:
    """
    Make a function that creates valid packings.

    :param inst: the two-dimensional bin packing instance
    :param random: the random number generator to use
    :returns: a function that can make packings valid
    """
    search_space = SignedPermutations(inst.get_standard_item_sequence())
    encoding = (ImprovedBottomLeftEncoding1 if random.integers(2) == 0
                else ImprovedBottomLeftEncoding2)(inst)
    op0 = Op0ShuffleAndFlip(search_space)

    def __make_valid(ra: rnd.Generator,
                     y: Packing, ss=search_space,
                     en=encoding, o0=op0) -> Packing:
        x = ss.create()
        o0.op0(ra, x)
        en.decode(x, y)
        return y

    return __make_valid


def make_packing_invalid(random: Generator = __RANDOM) \
        -> Callable[[Packing], Packing]:
    """
    Make a function that creates invalid packings.

    :param random: the random number generator to use
    :returns: a function that can make packings invalid
    """

    def __make_invalid(x: Packing, ri=random.integers) -> Packing:
        not_finished: bool = True
        while not_finished:
            while ri(2) == 0:
                x[ri(len(x)), ri(6)] = -1
                not_finished = False
            while ri(2) == 0:
                second = first = ri(len(x))
                while second == first:
                    second = ri(len(x))
                x[first, 1] = x[second, 1]
                x[first, 2] = x[second, 2] - 1
                x[first, 3] = x[second, 3] - 1
                x[first, 4] = x[second, 4] + 1
                x[first, 5] = x[second, 5] + 1
        return x

    return __make_invalid


def validate_algorithm_on_1_2dbinpacking(
        algorithm: Algorithm | Callable[
            [Instance, SignedPermutations, Objective], Algorithm],
        instance: str | None = None, max_fes: int = 100,
        random: Generator = __RANDOM) -> None:
    """
    Check the validity of a black-box algorithm on the 2d bin packing.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    :param random: the default random generator to use
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    search_space = SignedPermutations(inst.get_standard_item_sequence())
    solution_space = PackingSpace(inst)
    encoding = (ImprovedBottomLeftEncoding1 if random.integers(2) == 0
                else ImprovedBottomLeftEncoding2)(inst)
    objective = BinCountAndLastEmpty(inst)
    if callable(algorithm):
        algorithm = algorithm(inst, search_space, objective)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm, call=True)

    validate_algorithm(algorithm=algorithm,
                       solution_space=solution_space,
                       objective=objective,
                       search_space=search_space,
                       encoding=encoding,
                       max_fes=max_fes)


def validate_algorithm_on_2dbinpacking(
        algorithm: Callable[[Instance, SignedPermutations,
                             Objective], Algorithm],
        max_fes: int = 100, random: Generator = __RANDOM) -> None:
    """
    Validate an algorithm on a set of bin packing instances.

    :param algorithm: the algorithm factory
    :param max_fes: the maximum FEs
    :param random: the random number generator
    """
    for i in binpacking_instances_for_tests(random):
        validate_algorithm_on_1_2dbinpacking(algorithm, i, max_fes, random)


def validate_objective_on_1_2dbinpacking(
        objective: Objective | Callable[[Instance], Objective],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on 1 2D bin packing instance.

    :param objective: the objective function or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    if callable(objective):
        objective = objective(inst)

    validate_objective(
        objective=objective,
        solution_space=PackingSpace(inst),
        make_solution_space_element_valid=make_packing_valid(inst),
        is_deterministic=True)


def validate_objective_on_2dbinpacking(
        objective: Objective | Callable[[Instance], Objective],
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on bin packing instances.

    :param objective: the objective function or a factory creating it
    :param random: the random number generator
    """
    for i in binpacking_instances_for_tests(random):
        validate_objective_on_1_2dbinpacking(objective, i, random)


def validate_signed_permutation_encoding_on_1_2dbinpacking(
        encoding: Encoding | Callable[[Instance], Encoding],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate a signed permutation encoding on one 2D bin packing instance.

    :param encoding: the encoding or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    if callable(encoding):
        encoding = encoding(inst)
    inst = Instance.from_resource(instance)

    x_space = SignedPermutations(inst.get_standard_item_sequence())
    validate_space(x_space)

    y_space = PackingSpace(inst)
    validate_space(y_space, make_element_valid=None)

    validate_encoding(encoding, x_space, y_space)

    x = x_space.create()
    x_space.validate(x)

    y = y_space.create()
    encoding.decode(x, y)
    y_space.validate(y)

    random.shuffle(x)
    ri = random.integers
    for i, xx in enumerate(x):
        if ri(2) == 0:
            x[i] = -xx
    encoding.decode(x, y)
    y_space.validate(y)

    x_str = x_space.to_str(x)
    x_2 = x_space.from_str(x_str)
    if not x_space.is_equal(x, x_2):
        raise ValueError("error in space to/from_str")
    if not np.array_equal(x, x_2):
        raise ValueError("error in space to/from_str and is_equal")

    y_2 = y_space.create()
    encoding.decode(x_2, y_2)
    if not y_space.is_equal(y, y_2):
        raise ValueError("encoding is not deterministic")
    if not np.array_equal(y, y_2):
        raise ValueError(
            "encoding is not deterministic and error in space.is_equal")


def validate_signed_permutation_encoding_on_2dbinpacking(
        encoding: Encoding | Callable[[Instance], Encoding],
        random: Generator = __RANDOM) -> None:
    """
    Validate a signed permutation encoding function on bin packing instances.

    :param encoding: the encoding or a factory creating it
    :param random: the random number generator
    """
    for i in binpacking_instances_for_tests(random):
        validate_signed_permutation_encoding_on_1_2dbinpacking(
            encoding, i, random)



"""
The tour length objective function for tours in path representation.

A Traveling Salesperson Problem (TSP) instance  is defined as a
fully-connected graph with
:attr:`~moptipyapps.tsp.instance.Instance.n_cities` nodes. Each edge in the
graph has a weight, which identifies the distance between the nodes. The goal
is to find the *shortest* tour that visits every single node in the graph
exactly once and then returns back to its starting node. Then nodes are
usually called cities. In this file, we present methods for loading instances
of the TSP as distance matrices `A`. In other words, the value at `A[i, j]`
identifies the travel distance from `i` to `j`.

A tour can be represented in path representation, which means that it is
stored as a permutation of the numbers `0` to `n_cities-1`. The number at
index `k` identifies that `k`-th city to visit. So the first number in the
permutation identifies the first city, the second number the second city,
and so on.

The length of the tour can be computed by summing up the distances from the
`k`-th city to the `k+1`-st city, for `k` in `0..n_cities-2` and then adding
the distance from the last city to the first city. This is what the function
:func:`tour_length` is doing. This function is then wrapped as objective
function object in :class:`TourLength`.

Important initial work on this code has been contributed by Mr. Tianyu LIANG
(梁天宇), <liangty@stu.hfuu.edu.cn> a Master's student at the Institute of
Applied Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School
of Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei
University (合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the
supervision of Prof. Dr. Thomas Weise (汤卫思教授).

1. Gerhard Reinelt. TSPLIB - A Traveling Salesman Problem Library.
   *ORSA Journal on Computing* 3(4):376-384. November 1991.
   https://doi.org/10.1287/ijoc.3.4.376.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/
2. Gerhard Reinelt. *TSPLIB95.* Heidelberg, Germany: Universität
   Heidelberg, Institut für Angewandte Mathematik. 1995.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp95.pdf
3. David Lee Applegate, Robert E. Bixby, Vašek Chvátal, and William John Cook.
   *The Traveling Salesman Problem: A Computational Study.* Second Edition,
   2007. Princeton, NJ, USA: Princeton University Press. Volume 17 of
   Princeton Series in Applied Mathematics. ISBN: 0-691-12993-2.
4. Gregory Z. Gutin and Abraham P. Punnen. *The Traveling Salesman Problem and
   its Variations.* 2002. Kluwer Academic Publishers. Volume 12 of
   Combinatorial Optimization. https://doi.org/10.1007/b101971.
5. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   Sejla Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review*
   13(2):129-170. April 1999. https://doi.org/10.1023/A:1006529012972.
6. Eugene Leighton Lawler, Jan Karel Lenstra, Alexander Hendrik George Rinnooy
   Kan, and David B. Shmoys. *The Traveling Salesman Problem: A Guided Tour of
   Combinatorial Optimization.* September 1985. Chichester, West Sussex, UK:
   Wiley Interscience. In Estimation, Simulation, and
   Control - Wiley-Interscience Series in Discrete Mathematics and
   Optimization. ISBN: 0-471-90413-9
7. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, and Thomas Weise.
   Solving the Traveling Salesperson Problem using Frequency Fitness
   Assignment. In *Proceedings of the IEEE Symposium on Foundations of
   Computational Intelligence (IEEE FOCI'22),* part of the *IEEE Symposium
   Series on Computational Intelligence (SSCI'22),* December 4-7, 2022,
   Singapore. Pages 360-367. IEEE.
   https://doi.org/10.1109/SSCI51031.2022.10022296.
8. Thomas Weise, Raymond Chiong, Ke Tang, Jörg Lässig, Shigeyoshi Tsutsui,
   Wenxiang Chen, Zbigniew Michalewicz, and Xin Yao. Benchmarking Optimization
   Algorithms: An Open Source Framework for the Traveling Salesman Problem.
   *IEEE Computational Intelligence Magazine.* 9(3):40-52. August 2014.
   https://doi.org/10.1109/MCI.2014.2326101.
"""

from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from pycommons.types import type_error

from moptipyapps.shared import SCOPE_INSTANCE
from moptipyapps.tsp.instance import Instance


@numba.njit(cache=True, inline="always", fastmath=False, boundscheck=False)
def tour_length(instance: np.ndarray, x: np.ndarray) -> int:
    """
    Compute the length of a tour.

    :param instance: the distance matrix
    :param x: the tour
    :return: the length of the tour `x`
    """
    result: int = 0
    last: int = x[-1]
    for cur in x:
        result = result + instance[last, cur]
        last = cur
    return result


class TourLength(Objective):
    """The tour length objective function."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the tour length objective function.

        :param instance: the tsp instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the TSP instance we are trying to solve
        self.instance: Final[Instance] = instance

    def evaluate(self, x) -> int:
        """
        Compute the length of a tour in path representation.

        :param x: the tour in path representation
        :return: the tour length
        """
        return tour_length(self.instance, x)

    def lower_bound(self) -> int:
        """
        Get the lower bound of the tour length.

        :return: the lower bound of the tour length
        """
        return self.instance.tour_length_lower_bound

    def upper_bound(self) -> int:
        """
        Get the upper bound of the tour length.

        :return: the upper bound of the tour length
        """
        return self.instance.tour_length_upper_bound

    def __str__(self):
        """
        Get the name of this objective function: always "tourLength".

        :return: "tourLength"
        :retval "tourLength": always
        """
        return "tourLength"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)



"""
A set of pre-defined controllers.

In this module, we provide a set of pre-defined controllers taken from the
works of NOACK, CORNEJO MACEDA, LI, and SUN of the Harbin Institute of
Technology in Shenzhen, China (哈尔滨工业大学(深圳)). We ignore the
parameterizations offered in the original works and instead synthesize the
parameter values by ourselves.

1. Yuxiang LI (李宇翔). Jet Mixing Enhancement using Deep Reinforcement
   Learning (基于深度强化学习的射流混合增强控制). MSc Thesis. Harbin Institute
   of Technology in Shenzhen, China (哈尔滨工业大学(深圳)).
   January 2023.
2. Wei SUN (孙伟). Wake Control of 1-2-3 Fluidic Pinball using Deep
   Reinforcement Learning (基于深度强化学习方法的 1-2-3 流体弹球尾流控制).
   MSc Thesis. Harbin Institute of Technology in Shenzhen, China
   (哈尔滨工业大学(深圳)). January 2023.
3. Guy Yoslan CORNEJO MACEDA, François LUSSEYRAN, and Bernd R. NOACK.
   xMLC: A Toolkit for Machine Learning Control, First Edition.
   Machine Learning Tools in Fluid Mechanics, Vol 2.
   Shenzhen & Paris; Universitätsbibliothek der Technischen Universität
   Braunschweig. 2022 https://doi.org/10.24355/dbbs.084-202208220937-0
"""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cornejo_maceda(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a parameterized version of Cornejo Maceda's Law.

    See page 16, chapter 3.3.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    z = np.tanh(state[0] - state[1])
    b = params[0]
    z = np.tanh(1.0 if b == 0 else z / b)
    b = params[1]
    z = np.tanh(1.0 if b == 0 else z / b)
    b = params[2]
    out[0] = np.tanh(1.0 if b == 0 else z / b)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __table_3_1_ga(state: np.ndarray, _: float,
                   params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute the law evolved by genetic algorithm in Table 3-1.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    out[0] = state[0] * params[0] + state[1] * params[1]


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __table_3_1_lgpc(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute the law evolved by LGPC algorithm in Table 3-1.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    a: Final[float] = state[0] * params[0] + params[1]
    out[0] = params[2] * np.sin((params[3] / a) if (a != 0.0) else 1.0)


def predefined(system: System) -> tuple[Controller, ...]:
    """
    Create a set of pre-defined controllers for the given equations object.

    :param system: the equations object
    :return: the linear controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("cornejo_maceda", 2, 1, 3, __cornejo_maceda), )
    if system.state_dims == 3:
        return (Controller("table_3_1_ga", 3, 1, 2, __table_3_1_ga),
                Controller("table_3_1_lgpc", 3, 1, 4, __table_3_1_lgpc))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
Different encodings for the two-dimensional bin packing problem.

The following encodings are implemented:

- The improved bottom-left encoding 1,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1`, processes a
  permutation of objects from beginning to end and places the objects into
  the last bin (according to the improved-bottom-left method) until that bin
  is full and then begins to put them into the next bin.
- The improved bottom-left encoding 2,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_2`, processes a
  permutation of objects from beginning to end and places the objects into
  the first bin into which they fit (according to the improved-bottom-left
  method). It is slower than
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1` but its results
  are never worse for any permutation and better for several.
"""



"""
An objective function for minimizing the number of bins in packings.

This function returns the number of bins.
"""
from typing import Final

from moptipy.api.objective import Objective
from pycommons.types import type_error

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.packing import IDX_BIN

#: the name of the bin count objective function
BIN_COUNT_NAME: Final[str] = "binCount"


def ceil_div(a: int, b: int) -> int:
    """
    Compute a ceiling division.

    This function is needed by sub-classes.

    :param a: the number to be divided by `b`
    :param b: the number dividing `a`
    :return: the rounded-up result of the division

    >>> ceil_div(1, 1)
    1
    >>> ceil_div(98, 98)
    1
    >>> ceil_div(98, 99)
    1
    >>> ceil_div(98, 97)
    2
    >>> ceil_div(3, 1)
    3
    >>> ceil_div(3, 2)
    2
    >>> ceil_div(3, 3)
    1
    >>> ceil_div(3, 4)
    1
    >>> ceil_div(4, 1)
    4
    >>> ceil_div(4, 2)
    2
    >>> ceil_div(4, 3)
    2
    >>> ceil_div(4, 4)
    1
    >>> ceil_div(4, 5)
    1
    >>> ceil_div(4, 23242398)
    1
    """
    return -((-a) // b)


class BinCount(Objective):
    """Compute the number of bins."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the number of bins objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the internal instance reference
        self._instance: Final[Instance] = instance

    def evaluate(self, x) -> int:
        """
        Get the number of bins.

        :param x: the packing
        :return: the number of bins used
        """
        return int(x[:, IDX_BIN].max())

    def lower_bound(self) -> int:
        """
        Get the lower bound of the number of bins objective.

        :return: the lower bound for the number of required bins, i.e.,
            :attr:`~moptipyapps.binpacking2d.instance.Instance.\
lower_bound_bins`

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.lower_bound_bins
        1
        >>> BinCount(ins).lower_bound()
        1

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.lower_bound_bins
        2
        >>> BinCount(ins).lower_bound()
        2

        >>> ins = Instance("c", 10, 50, [[10, 5, 20], [30, 3, 10], [5, 5, 1]])
        >>> ins.lower_bound_bins
        4
        >>> BinCount(ins).lower_bound()
        4
        """
        return self._instance.lower_bound_bins

    def is_always_integer(self) -> bool:
        """
        Return `True` because there are only integer bins.

        :retval True: always
        """
        return True

    def upper_bound(self) -> int:
        """
        Get the upper bound of the number of bins plus emptiness.

        :return: the number of items in the instance, i.e.,
            :attr:`~moptipyapps.binpacking2d.instance.Instance.n_items`

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        3
        >>> BinCount(ins).upper_bound()
        3

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        12
        >>> BinCount(ins).upper_bound()
        12

        >>> ins = Instance("c", 10, 50, [[10, 5, 20], [30, 3, 10], [5, 5, 1]])
        >>> ins.n_items
        31
        >>> BinCount(ins).upper_bound()
        31
        """
        return self._instance.n_items

    def to_bin_count(self, z: int) -> int:
        """
        Get the bin count corresponding to an objective value.

        :param z:
        :return: the value itself
        """
        return z

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCount`
        :retval "binCount": always
        """
        return BIN_COUNT_NAME



"""
An objective function that can be used to synthesize system models.

We consider the dynamic system to be a function `D(s, c) = ds/dt`, where

- `s` is the state vector (e.g., two-dimensional for the Stuart-Landau system,
  see :mod:`~moptipyapps.dynamic_control.systems.stuart_landau`, and
  three-dimensional for the Lorenz system, see
  :mod:`~moptipyapps.dynamic_control.systems.lorenz`),
- `c` is the control vector (one-dimensional in both cases), and
- `ds/dt` is the state differential (again two respectively three-dimensional).

The :class:`~moptipyapps.dynamic_control.objective.FigureOfMerit` objective
function allows us to collect tuples of `(s,c)` and `ds/dt` vectors. So all we
need to do is to train a model `M` that receives as input a vector `x=(s,c)`
(where `(s,c)` be the concatenation of `s` and `c`) and fill as output a vector
`ds/dt`.

The objective function in this module minimizes the root mean square error
over the model-computed `ds/dt` vectors and the actual `ds/dt` vectors.
The model objective function is used by the
:mod:`~moptipyapps.dynamic_control.surrogate_optimizer` algorithm.
"""
from typing import Callable, Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_FLOAT

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.objective import FigureOfMerit


@numba.njit(cache=False, inline="always", fastmath=True, boundscheck=False)
def _evaluate(x: np.ndarray, pin: np.ndarray, pout: np.ndarray,
              temp_1: np.ndarray, temp_2: np.ndarray,
              eq: Callable[[np.ndarray, float, np.ndarray,
                            np.ndarray], None]) -> float:
    """
    Compute the RMSE differences between expected and actual model output.

    :param x: the model parameterization
    :param pin: the input vectors
    :param pout: the expected output vectors, flattened
    :param temp_1: the long temporary array to receive the result values
    :param temp_2: the temporary array to receive the state output
    :param eq: the equations
    :return: the mean of the `log(x+1)` of the squared differences `x`

    >>> vx = np.array([0.5, 0.2], float)
    >>> vpin = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]], float)
    >>> vpout = np.array([[10, 5], [14, 9], [18, 8]], float)
    >>> vtemp_1 = np.empty(3, float)
    >>> vtemp_2 = np.empty(2, float)
    >>> @numba.njit(cache=False, inline="always", fastmath=True)
    ... def _func(lstate: np.ndarray, _: float, lx: np.ndarray,
    ...           lout: np.ndarray) -> None:
    ...     lout[:] = (lstate[0:2] * lx[0]) - lx[1] - lstate[-1]
    >>> _evaluate(vx, vpin, vpout, vtemp_1, vtemp_2, _func)
    22.680823177337874
    >>> y_1_1 = ((0 * 0.5) - 0.2 - 3) - 10
    >>> y_1_2 = ((1 * 0.5) - 0.2 - 3) - 5
    >>> r_1 = y_1_1 ** 2 + y_1_2 ** 2
    >>> y_2_1 = ((4 * 0.5) - 0.2 - 7) - 14
    >>> y_2_2 = ((5 * 0.5) - 0.2 - 7) - 9
    >>> r_2 = y_2_1 ** 2 + y_2_2 ** 2
    >>> y_3_1 = ((8 * 0.5) - 0.2 - 11) - 18
    >>> y_3_2 = ((9 * 0.5) - 0.2 - 11) - 8
    >>> r_3 = y_3_1 ** 2 + y_3_2 ** 2
    >>> np.sqrt(np.array([r_1, r_2, r_3])).mean()
    22.680823177337874
    """
    for i, row in enumerate(pin):  # iterate over all row=(s, c) tuples
        eq(row, 0.0, x, temp_2)  # store the equation results
        temp_1[i] = np.square(np.subtract(temp_2, pout[i], temp_2),
                              temp_2).sum()
    return np.sqrt(temp_1, temp_1).mean()


class ModelObjective(Objective):
    """
    The objective for modeling.

    This objective function works on sequences of tuples `(s, c)` of the
    system state `s` and controller output `c` as well as the corresponding
    `ds/dt` differentials. The goal is to train a model `M(s, c, q)` that will
    compute the `ds/dt` values reasonably accurately. The model is
    parameterized with some vector `q`, think of `M` being, e.g., an
    artificial neural network and `q` being its weight vector. To find good
    values of `q`, this objective here computes the squared differences
    between the values `M(s, c, q)` and the expected outputs `ds/dt`.

    These squared differences are then either averaged directly or the
    `expm1(mean(logp1(...))))` hack of the
    :class:`~moptipyapps.dynamic_control.objective.FigureOfMeritLE` is used to
    alleviate the impact of large differentials, depending on which "real"
    (controller-synthesis) objective is passed into the constructor).

    The tuples `(s, c)` and `ds/dt` are pulled from the real objective with
    the method :meth:`~moptipyapps.dynamic_control.model_objective.\
ModelObjective.begin`. The `ds/dt` rows are flattened for performance reasons.
    """

    def __init__(self, real: FigureOfMerit,
                 model: Controller) -> None:
        """
        Create a model objective compatible to the given figure of merit.

        :param real: the objective used for the real optimization problem
        :param model: the model
        """
        #: the equations of the model
        self.__equations: Callable[[np.ndarray, float, np.ndarray,
                                    np.ndarray], None] = model.controller
        #: the result
        self.__temp_1: np.ndarray | None = None
        #: the input
        self.__in: np.ndarray | None = None
        #: the output
        self.__out: np.ndarray | None = None
        #: the differentials getter
        self.__get_differentials: Final[Callable[[], tuple[
            np.ndarray, np.ndarray]]] = real.get_differentials
        #: the temporary array
        self.__temp_2: Final[np.ndarray] = np.empty(
            real.instance.system.state_dims, DEFAULT_FLOAT)
        #: the real figure of merit name
        self.__real_name: Final[str] = str(real)

    def begin(self) -> None:
        """
        Begin a model optimization run.

        This function pulls the training data from the actual
        controller-synthesis objective function, which is an instance of
        :class:`~moptipyapps.dynamic_control.objective.FigureOfMerit`, via the
        method :meth:`~moptipyapps.dynamic_control.objective.FigureOfMerit.\
get_differentials` and allocates internal data structures accordingly.
        """
        self.__in, self.__out = self.__get_differentials()
        self.__temp_1 = np.empty(len(self.__out), self.__out.dtype)

    def evaluate(self, x: np.ndarray) -> float:
        """
        Evaluate a model parameterization.

        :param x: the model parameterization
        :return: the objective value
        """
        return _evaluate(x, self.__in, self.__out, self.__temp_1,
                         self.__temp_2, self.__equations)

    def end(self) -> None:
        """End a model optimization run and free the associated memory."""
        self.__temp_1 = None
        self.__in = None
        self.__out = None

    def __str__(self) -> str:
        """
        Get the name of this objective.

        :return: the name of this objective
        """
        return "modelRMSE"

    def lower_bound(self) -> float:
        """
        Get the lower bound of the model objective, which is `0`.

        :returns: 0.0
        """
        return 0.0

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("figureOfMeritName", self.__real_name)
        logger.key_value("nSamples",
                         0 if self.__in is None else len(self.__in))
